#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **dictionary mutators** (i.e., low-level callables modifying the
contents of passed dictionaries in various general-purpose ways).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilMappingException
from beartype.typing import (
    AbstractSet,
    Collection,
    Sequence,
)
from beartype._util.text.utiltextrepr import represent_object
from collections.abc import (
    Hashable,
    Sequence as SequenceABC,
    Mapping,
    MutableMapping,
    Set,  # <-- equivalent to "typing.AbstractSet", interestingly
)

# ....................{ MERGERS                            }....................
def merge_mappings(*mappings: Mapping) -> Mapping:
    '''
    Safely merge all passed mappings if these mappings contain no **key-value
    collisions** (i.e., if these mappings either contain different keys *or*
    share one or more key-value pairs) *or* raise an exception otherwise (i.e.,
    if these mappings contain one or more key-value collisions).

    Since this function only safely merges mappings and thus *never* silently
    overrides any key-value pair of either mapping, order is insignificant;
    this function returns the same mapping regardless of the order in which
    these mappings are passed.

    Caveats
    -------
    This function creates and returns a new mapping of the same type as that of
    the first mapping. That type *must* define an ``__init__()`` method with
    the same signature as the standard :class:`dict` type; if this is *not* the
    case, an exception is raised.

    Parameters
    ----------
    mappings: tuple[Mapping]
        Tuple of two or more mappings to be safely merged.

    Returns
    -------
    Mapping
        Mapping of the same type as that of the first mapping created by safely
        merging these mappings.

    Raises
    ------
    _BeartypeUtilMappingException
        If either:

        * No mappings are passed.
        * Only one mappings are passed.
        * Two or more mappings are passed, but these mappings contain one or
          more key-value collisions.

    See Also
    --------
    :func:`.die_if_mappings_two_items_collide`
        Further details.
    '''

    # Return either...
    return (
        # If only two mappings are passed, defer to a function optimized for
        # merging two mappings.
        merge_mappings_two(mappings[0], mappings[1])
        if len(mappings) == 2 else
        # Else, three or more mappings are passed. In this case, defer to a
        # function optimized for merging three or more mappings.
        merge_mappings_two_or_more(mappings)
    )


def merge_mappings_two(mapping_a: Mapping, mapping_b: Mapping) -> Mapping:
    '''
    Safely merge the two passed mappings if these mappings contain no key-value
    collisions *or* raise an exception otherwise.

    Parameters
    ----------
    mapping_a: Mapping
        First mapping to be merged.
    mapping_b: Mapping
        Second mapping to be merged.

    Returns
    -------
    Mapping
        Mapping of the same type as that of the first mapping created by safely
        merging these mappings.

    Raises
    ------
    _BeartypeUtilMappingException
        If these mappings contain one or more key-value collisions.

    See Also
    --------
    :func:`beartype._util.kind.maplike.utilmaptest.die_if_mappings_two_items_collide`
        Further details.
    '''

    # If the first mapping is empty, return the second mapping as is.
    if not mapping_a:
        return mapping_b
    # Else, the first mapping is non-empty.
    #
    # If the second mapping is empty, return the first mapping as is.
    elif not mapping_b:
        return mapping_a
    # Else, both mappings are non-empty.

    # Avoid circular import dependencies.
    from beartype._util.kind.maplike.utilmaptest import (
        die_if_mappings_two_items_collide)

    # If these mappings contain a key-value collision, raise an exception.
    die_if_mappings_two_items_collide(mapping_a, mapping_b)
    # Else, these mappings contain *NO* key-value collisions.

    # Merge these mappings. Note that:
    # * No unsafe collisions exist. Ergo, the order in which these mappings are
    #   merged is irrelevant.
    # * The active Python interpreter targets Python >= 3.9 and thus supports
    #   "PEP 584 -- Add Union Operators To dict". Ergo, these mappings are
    #   optimally mappable with the faster and terser dict union operator.
    return mapping_a | mapping_b  # type: ignore[operator]


def merge_mappings_two_or_more(mappings: Sequence[Mapping]) -> Mapping:
    '''
    Safely merge the one or more passed mappings if these mappings contain no
    key-value collisions *or* raise an exception otherwise.

    Parameters
    ----------
    mappings: SequenceABC[Mapping]
        SequenceABC of two or more mappings to be safely merged.

    Returns
    -------
    Mapping
        Mapping of the same type as that of the first mapping created by safely
        merging these mappings.

    Raises
    ------
    _BeartypeUtilMappingException
        If these mappings contain one or more key-value collisions.

    See Also
    --------
    :func:`beartype._util.kind.maplike.utilmaptest.die_if_mappings_two_items_collide`
        Further details.
    '''
    assert isinstance(mappings, SequenceABC), f'{repr(mappings)} not sequence.'

    # Number of passed mappings.
    MAPPINGS_LEN = len(mappings)

    # If less than two mappings are passed, raise an exception.
    if MAPPINGS_LEN < 2:
        # If only one mapping is passed, raise an appropriate exception.
        if MAPPINGS_LEN == 1:
            raise _BeartypeUtilMappingException(
                f'Two or more mappings expected, but only one mapping '
                f'{represent_object(mappings[0])} passed.')
        # Else, no mappings are passed. Raise an appropriate exception.
        else:
            raise _BeartypeUtilMappingException(
                'Two or more mappings expected, but no mappings passed.')
    # Else, two or more mappings are passed.
    assert isinstance(mappings[0], Mapping), (
        f'First mapping {repr(mappings[0])} not mapping.')

    # Merged mapping to be returned, initialized to the merger of the first two
    # passed mappings.
    mapping_merged = merge_mappings_two(mappings[0], mappings[1])

    # If three or more mappings are passed...
    if MAPPINGS_LEN > 2:
        # For each of the remaining mappings...
        for mapping in mappings[2:]:
            # Merge this mapping into the merged mapping to be returned.
            mapping_merged = merge_mappings_two(mapping_merged, mapping)
    # Else, only two mappings are passed. In these case, these mappings have
    # already been merged above.

    # Return this merged mapping.
    return mapping_merged

# ....................{ REMOVERS                           }....................
def remove_mapping_keys(mapping: MutableMapping, keys: AbstractSet) -> None:
    '''
    Safely remove all keys from the passed mapping that are items of the passed
    set, silently ignoring any keys that are *not* already in this mapping.

    This method is thread-safe against concurrent mutation by competing threads.

    Parameters
    ----------
    mapping : MutableMapping
        Mapping to remove these keys from.
    keys : AbstractSet
        Set of all keys to be removed.
    '''
    assert isinstance(mapping, MutableMapping), (
        f'{repr(mapping)} not mutable mapping.')
    assert isinstance(keys, Set), f'{repr(keys)} not set.'

    # Set of all existing keys to be removed from this mapping, efficiently
    # defined as the intersection of the keys of this mapping with the passed
    # set of keys.
    mapping_keys = mapping.keys() & keys

    # If this mapping contains *NONE* of these keys, silently reduce to a noop.
    #
    # Note that this is an optimization. However, "for" loops internally raise
    # "StopIteration" exceptions on halt and thus incur a non-trivial cost.
    # Ergo, this is a non-trivial (read: valuable) optimization.
    if not mapping_keys:
        return
    # Else, this mapping contains one or more of these keys.

    # dict.pop() method bound to this mapping, localized for efficiency.
    mapping_pop = mapping.pop

    #FIXME: [SPEED] Optimize into a "while" loop, please. *sigh*
    # For each of these existing keys to be removed...
    #
    # Note that CPython currently provides *NO* efficient means of removing
    # multiple keys from a mapping in a single expression. This is known to be
    # the best approach, but it still requires manual iteration over these keys.
    for mapping_key in mapping_keys:
        # Safely remove the key-value pair from this mapping whose key is this
        # key if this mapping contains this key *OR* silently reduce to a noop
        # otherwise (i.e., if this mapping does *NOT* contain this key).
        #
        # Note that:
        # * By the prior validation, this key *SHOULD* be guaranteed to exist in
        #   this mapping. However, CPython now allows end users to externally
        #   disable the Global Interpreter Lock (GIL). If an external thread is
        #   concurrently mutating this mapping, this guarantee of key existence
        #   no longer holds. Since the dict.__del__() and dict.pop() methods
        #   mostly exhibit similar performance, there is *NO* practical benefit
        #   to increasing the fragility of this function by adopting the
        #   dict.__del__() dunder method over the dict.pop() method. Ergo, we
        #   prefer this least fragile approach.
        # * This is the standard idiom for efficiently removing key-value pairs
        #   where this key is *NOT* guaranteed to exist in this mapping.
        #   Simplicity and speed supercedes readability, sadly.
        mapping_pop(mapping_key, None)

# ....................{ UPDATERS                           }....................
def update_mapping(mapping_trg: MutableMapping, mapping_src: Mapping) -> None:
    '''
    Safely update in-place the first passed mapping with all key-value pairs of
    the second passed mapping if these mappings contain no **key-value
    collisions** (i.e., if these mappings either only contain different keys
    *or* share one or more key-value pairs) *or* raise an exception otherwise
    (i.e., if these mappings contain one or more of the same keys associated
    with different values).

    Parameters
    ----------
    mapping_trg: MutableMapping
        Target mapping to be safely updated in-place with all key-value pairs
        of ``mapping_src``. This mapping is modified by this function and
        *must* thus be mutable.
    mapping_src: Mapping
        Source mapping to be safely merged into ``mapping_trg``. This mapping
        is *not* modified by this function and may thus be immutable.

    Raises
    ------
    _BeartypeUtilMappingException
        If these mappings contain one or more key-value collisions.

    See Also
    --------
    :func:`beartype._util.kind.maplike.utilmaptest.die_if_mappings_two_items_collide`
        Further details.
    '''
    assert isinstance(mapping_trg, MutableMapping), (
        f'{repr(mapping_trg)} not mutable mapping.')
    assert isinstance(mapping_src, Mapping), (
        f'{repr(mapping_src)} not mapping.')

    # If the second mapping is empty, silently reduce to a noop.
    if not mapping_src:
        return
    # Else, the second mapping is non-empty.

    # Avoid circular import dependencies.
    from beartype._util.kind.maplike.utilmaptest import (
        die_if_mappings_two_items_collide)

    # If these mappings contain a key-value collision, raise an exception.
    die_if_mappings_two_items_collide(mapping_trg, mapping_src)
    # Else, these mappings contain *NO* key-value collisions.

    # Update the former mapping from the latter mapping. Since no unsafe
    # collisions exist, this update is now guaranteed to be safe.
    mapping_trg.update(mapping_src)


def update_mapping_keys(
    # Mandatory parameters.
    mapping: MutableMapping,
    keys: Collection[Hashable],

    # Optional parameters.
    value: object = None,
) -> None:
    '''
    Map the passed keys to the same passed value in the passed mapping.

    This function is an efficient alternative to the standard idiom for adding
    multiple keys to a mapping when those keys *all* share the same value:

    .. code-block:: python

       # This function effectively does this -- except *WAY* faster. "for" loops
       # are excrutiatingly slow, due to raising "StopIteration" exceptions on
       # normal loop termination.
       for key in keys:
            mapping[key] = value

    This function is principally useful for efficiently creating and maintaining
    an **insertion-order mimic set** (i.e., a set-like object that preserves
    insertion order despite not actually being a set). While neither the builtin
    :class:`set` nor :class:`frozenset` types preserve insertion order, the
    builtin :class:`dict` type *does*. An insertion-order mimic set can thus be
    trivially constructed as a dictionary whose:

    * Keys are the desired ordered-preserving set members.
    * Values are ignorable (e.g., :data:`None`).

    Parameters
    ----------
    mapping : MutableMapping
        Mapping to add these key-value pairs to.
    keys : Iterable[Hashable]
        Collection of all keys to be added to this mapping.
    value : object, optional
        Singular value to map *all* of these keys to. Defaults to :data:`None`
        to trivialize insertion-order sets.
    '''
    assert isinstance(mapping, MutableMapping), (
        f'{repr(mapping)} not mutable mapping.')

    # Efficiently map *ALL* of these keys to this same value in this mapping.
    # Dismantled, this is:
    # * "(value,) * len(keys)", a tuple of the same length as the passed
    #   collection of keys whose items are *ALL* the passed value repeated
    #   verbatim. Since Python optimizes tuple creation, this operation is
    #   highly efficient despite superficially appearing to be insane.
    # * "zip(...)", an iterable pairing each of these keys with each of these
    #   duplicate values. If this iterable was itself a tuple, its contents
    #   would resemble:
    #       ((keys[0], value), (keys[1], value), ..., (keys[N], value))
    #
    # Note that this logic also implicitly handles the edge case of the caller
    # passing an empty iterable of keys. Why? Because the zip() builtin
    # dynamically generates an iterable whose length is the minimum of the
    # lengths of the passed iterables. If one of the passed iterables is empty,
    # the iterable that zip() generates will also be empty. Hallelujah, Python!
    mapping.update(zip(keys, (value,) * len(keys)))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hookable FastMCP integration submodule** (i.e., data module
defining FastMCP-specific runnables decorated by the third-party
decorator-hostile :func:`fastmcp.FastMCP.tool` decorator which the
:mod:`beartype.beartype` decorator will then be injected *after* rather than
*before* as the earliest decorator, mimicking real-world usage of
:func:`beartype.claw` import hooks from external callers).

See Also
--------
https://github.com/beartype/beartype/issues/540
    StackOverflow issue exercised by this submodule.
https://gofastmcp.com/clients/client
    Official FastMCP documentation lightly inspiring this integration test.
'''

# ....................{ IMPORTS                            }....................
from fastmcp import (
    Client,
    FastMCP,
)

# ....................{ GLOBALS                            }....................
fastmcp_server = FastMCP('With stride colossal, on from hall to hall;')
'''
Arbitrary in-memory FastMCP server leveraging the default STDIO (i.e., standard
input and output) transport protocol.
'''


fastmcp_client = Client(fastmcp_server)
'''
Arbitrary in-memory FastMCP client connecting to this server of that protocol.
'''

# ....................{ FUNCTIONS                          }....................
@fastmcp_server.tool
def with_stride_colossal(on_from_hall_to_hall: str) -> int:
    '''
    Arbitrary function trivially satisfying FastMCP's runnable protocol by
    accepting an arbitrary string of mock input data and returning a byte string
    mocking the output text response of a model context protocol (MCP) passed
    that input.

    The :meth:`FastMCP.tool` decorator method wraps this function in a
    FastMCP-specific runnable object defining a ``run()`` method, which we then
    call below to validate this wrapping.
    '''

    # Return the trivial length of this string. We initially attempted to coerce
    # this string into a byte string, but Pydantic promptly complained about
    # that. Lol. "Y U H8 byte string, M8?"
    return len(on_from_hall_to_hall)

# ....................{ MAIN                               }....................
async def data_claw_fastmcp_main() -> None:
    '''
    Asynchronous coroutine performing all FastMCP-specific integration tests.

    Note that:

    * FastMCP implicitly transforms *all* FastMCP-decorated synchronous
      callables into asynchronous callables. Client FastMCP operations, in
      particular, *must* then be called "using the ``async with`` context
      manager for proper connection lifecycle management."
    * The caller *must* manually run this coroutine within an ``asyncio`` loop
      explicitly managed by the caller.
    '''
    # print('[data_claw_fastmcp] with_stride_colossal:')
    # print(with_stride_colossal.__module__)
    # print(with_stride_colossal.__class__.__name__)

    # ....................{ FASTMCP                        }....................
    # Inside a FastMCP-specific asynchronous context manager encapsulating this
    # client-server connection...
    async with fastmcp_client:
        # Implicitly assert the server connected to this client to be awake by
        # trivially pinging the server from this client.
        await fastmcp_client.ping()

        # ....................{ PASS                       }....................
        # Valid input to be passed to the FastMCP tool defined above.
        TOOL_INPUT = 'While far within each aisle and deep recess,'

        # Synchronously call the asynchronous FastMCP server tool defined above
        # with valid input. Doing so implicitly asserts that the import hook
        # registered by the caller respected the decorator-hostile
        # @fastmcp_server.tool method decorating the function defined above by
        # injecting the @beartype decorator as the last decorator on that
        # function.
        #
        # If this call raises an unexpected exception, then that import hook
        # failed to respect that the decorator-hostile @fastmcp_server.tool
        # decorator by instead injecting the @beartype decorator as the first
        # decorator on that function. Since @fastmcp_server.tool is
        # decorator-hostile and thus hostile to @beartype as well,
        # @fastmcp_server.tool prohibits *ANY* decorator from being applied
        # after itself.
        #
        # Note that:
        # * The object returned by this asynchronous client invocation of a
        #   FastMCP tool is a FastMCP-specific "CallToolResult" object. Look. We
        #   don't know either. This API is hell to test. Sanity now! See also:
        #     https://gofastmcp.com/clients/tools#handling-results
        # * There's little point in attempting to validate that this tool raises
        #   @beartype-specific type-checking violations. Why? Because it
        #   doesn't. FastMCP validates input through Pydantic *BEFORE*
        #   @beartype-specific type-checking can get there. It is what it is.
        tool_output = await fastmcp_client.call_tool(
            'with_stride_colossal', {'on_from_hall_to_hall': TOOL_INPUT})

        # Explicitly assert that call returned the expected output.
        # print(f'tool_output: {tool_output}')
        assert tool_output.data == len(TOOL_INPUT)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator **type-checking function code magic** (i.e., global string
constants embedded in the implementations of functions type-checking arbitrary
objects against arbitrary PEP-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ NAMES                              }....................
NAME_PREFIX = '__beartype_'
'''
Substring prefixing the names of all *other* global string constants declared by
this submodule.
'''

# ....................{ NAMES ~ func                       }....................
FUNC_CHECKER_NAME_PREFIX = f'{NAME_PREFIX}checker_'
'''
Substring prefixing the unqualified basenames of all type-checking raiser and
tester functions created by the
:func:`beartype._check.checkmake._make_func_checker` factory function.
'''

# ....................{ NAMES ~ parameter                  }....................
# To avoid colliding with the names of arbitrary caller-defined parameters, the
# beartype-specific hidden parameter names *MUST* be prefixed by "__beartype_".

ARG_NAME_ARGS_NAME_KEYWORDABLE = f'{NAME_PREFIX}args_name_keywordable'
'''
Name of the **private keywordable parameter name set** (i.e.,
:mod:`beartype`-specific hidden parameter whose default value is the frozen set
of the names of all parameters that either may *or* must be passed by keyword,
required to type-check a :func:`beartype.beartype`-decorated callable accepting
an annotated variadic keyword parameter like ``**kwargs: int``).
'''


ARG_NAME_CHECK_META = f'{NAME_PREFIX}check_meta'
'''
Name of the **private beartype type-checking call metadata** (i.e.,
:mod:`beartype`-specific hidden parameter whose default value is the
:class:`beartype._check.metadata.metacheck.BeartypeCheckMeta` dataclass instance
encapsulating *all* metadata required by each call to the wrapper function
type-checking a :func:`beartype.beartype`-decorated callable).
'''


ARG_NAME_CONF = f'{NAME_PREFIX}conf'
'''
Name of the **private beartype configuration parameter** (i.e.,
:mod:`beartype`-specific hidden parameter whose default value is the
:class:`beartype.BeartypeConf` instance configuring each wrapper function
generated by the :func:`beartype.beartype` decorator).
'''


ARG_NAME_EXCEPTION_PREFIX = f'{NAME_PREFIX}exception_prefix'
'''
Name of the **private exception prefix parameter** (i.e.,
:mod:`beartype`-specific hidden parameter whose default value is the human-readable
label prefixing the representation of the currently type-checked object in
exception messages raised when this object violates its type hint, conditionally
passed to wrappers generated by the :func:`beartype.door.die_if_unbearable`
type-checker injected for :pep:`526`-compliant annotated variable assignments by
:mod:`beartype.claw`-published import hooks).
'''


#FIXME: *REDUNDANT.* The same metadata is now directly accessible via the
#existing "{ARG_NAME_CHECK_META}.func" field available to *ALL* type-checking
#wrapper functions. Obsolete this redundant hidden parameter, please. *sigh*
ARG_NAME_FUNC = f'{NAME_PREFIX}func'
'''
Name of the **private decorated callable parameter** (i.e.,
:mod:`beartype`-specific hidden parameter whose default value is the decorated
callable passed to each wrapper function generated by the
:func:`beartype.beartype` decorator).
'''


ARG_NAME_GETRANDBITS = f'{NAME_PREFIX}getrandbits'
'''
Name of the **private getrandbits parameter** (i.e., :mod:`beartype`-specific
parameter whose default value is the highly performant C-based
:func:`random.getrandbits` function conditionally passed to wrappers generated
by the :func:`beartype.beartype` decorator whose type-checking logic requires
one or more random integers).
'''


ARG_NAME_GET_VIOLATION = f'{NAME_PREFIX}get_violation'
'''
Name of the **private exception raising parameter** (i.e.,
:mod:`beartype`-specific hidden parameter whose default value is the
:func:`beartype._check.error.errmain.get_func_pith_violation`
function raising human-readable exceptions on call-time type-checking failures
passed to each wrapper function generated by the :func:`beartype.beartype`
decorator).
'''


ARG_NAME_HINT = f'{NAME_PREFIX}hint'
'''
Name of the **private type hint parameter** (i.e., :mod:`beartype`-specific
parameter whose default value is the user-defined type hint unconditionally
passed to the current wrapper function generated by the
:func:`beartype.door.die_if_unbearable` type-checker receiving that hint).
'''


ARG_NAME_WARN = f'{NAME_PREFIX}warn'
'''
Name of the **standard warn function** (i.e., :mod:`beartype`-specific
parameter whose default value is the :func:`warnings.warn` function
conditionally passed to every wrapper function generated by the
:func:`beartype.beartype` decorator configured by either the
:attr:`beartype.BeartypeConf.violation_param_type` or
:attr:`beartype.BeartypeConf.violation_return_type` options to emit
non-fatal warnings rather than raise fatal exceptions).
'''

# ....................{ NAMES ~ var                        }....................
VAR_NAME_ARGS_LEN = f'{NAME_PREFIX}args_len'
'''
Name of the local variable providing the **positional argument count** (i.e.,
number of positional arguments passed to the current call).
'''


VAR_NAME_RANDOM_INT = f'{NAME_PREFIX}random_int'
'''
Name of the local variable providing a **pseudo-random integer** (i.e.,
unsigned 32-bit integer pseudo-randomly generated for subsequent use in
type-checking randomly indexed container items by the current call).
'''


VAR_NAME_VIOLATION = f'{NAME_PREFIX}violation'
'''
Name of the local variable providing the **violation exception** (i.e.,
exception describing a type-checking violation to be either raised as a fatal
exception or emitted as a non-fatal warning by the current call as configured by
the :attr:`beartype.BeartypeConf.violation_param_type` and
:attr:`beartype.BeartypeConf.violation_return_type` options).
'''

# ....................{ NAMES ~ var : pith                 }....................
VAR_NAME_PITH_PREFIX = f'{NAME_PREFIX}pith_'
'''
Substring prefixing all local variables providing a **pith** (i.e., either the
current parameter or return value *or* item contained in the current parameter
or return value type-checked by the current call).
'''


VAR_NAME_PITH_ROOT = f'{VAR_NAME_PITH_PREFIX}0'
'''
Name of the local variable providing the **root pith** (i.e., value of the
current parameter or return value being type-checked by the current call).
'''

# ....................{ CODE ~ pith                        }....................
CODE_PITH_ROOT_NAME_PLACEHOLDER = '?|PITH_ROOT_NAME`^'
'''
Placeholder source substring to be globally replaced by the **root pith name**
(i.e., name of the current parameter if called by the
:func:`pep_code_check_param` function *or* ``return`` if called by the
:func:`pep_code_check_return` function) in the parameter- and return-agnostic
code generated by the memoized
:func:`beartype._check.checkmake.make_code_raiser_func_pith_check` function.

See Also
--------
:attr:`beartype._data.error.dataerrmagic.EXCEPTION_PLACEHOLDER`
    Related commentary.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hook module caches** (i.e., private dictionary singletons
enabling relevant metadata including beartype configurations associated with
submodules of all packages previously registered in our global package trie to
be efficiently stored and retrieved based on various criteria).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.claw.dataclawmagic import BEARTYPE_OPTIMIZATION_MARKER
from beartype.roar import BeartypeClawImportConfException
from beartype.typing import Dict
from beartype._conf.confmain import BeartypeConf
from pprint import pformat

# Original cache_from_source() function defined by the private (*gulp*)
# "importlib._bootstrap_external" submodule, preserved *BEFORE* temporarily
# replacing that function with our beartype-specific variant in the
# "beartype.claw._importlib._clawimpload" submodule.
from importlib.util import cache_from_source as cache_from_source_original

# ....................{ SUBCLASSES                         }....................
class ModuleNameToBeartypeConf(Dict[str, 'BeartypeConf']):
    '''
    Non-thread-safe **hooked module beartype configuration cache** (i.e.,
    dictionary mapping from the fully-qualified name of each previously imported
    submodule of each package previously registered in our global package trie
    to the beartype configuration configuring type-checking by the
    :func:`beartype.beartype` decorator of that submodule).

    This dictionary subclass improves the human readability of exceptions raised
    by dunder methods of the :class:`dict` superclass (e.g., the
    :meth:`dict.__getitem__` dunder method), whose C-based implementations
    raise non-human-readable exceptions in common use cases encountered by end
    users leveraging beartype import hooks: e.g.,

    .. code-block:: python

        # Otherwise syntactically and semantically correct PEP 526-compliant
        # annotated assignment expressions like this previously raised spurious
        # non-human-readable exceptions from this dictionary resembling:
        #     KeyError: 'muh_module'  # <-- what does this even mean!?!?
        loves_philosophy: float = len('The fountains mingle with the river')

    Motivation
    ----------
    This cache provides an efficient :math:`O(1)` alternative to the
    comparatively less efficient
    :func:`beartype.claw._package.clawpkgtrie.get_package_conf_or_none` function,
    which exhibits worst-case runtime complexity of ``O(k)`` for ``k`` the
    maximum depth of our global package trie. Doing so enables the
    :mod:`beartype.claw._ast.clawastmain` submodule implementing our abstract
    syntax tree (AST) node transformer to trivially inject efficient code
    looking up the current beartype configuration associated with the currently
    transformed module into the body of that module, which would otherwise be
    quite non-trivial.

    Caveats
    -------
    **This cache is non-thread-safe.** The caller is responsible for
    guaranteeing thread-safety on writes to this cache. However, Note that reads
    of this cache are implicitly thread-safe. The :meth:`BeartypeConf.__new__`
    instantiator thread-safely stores strong references to the currently
    instantiated beartype configuration in both this and other caches. Since
    these caches and thus *all* configurations persist for the lifetime of the
    active Python interpreter, reads are effectively thread-safe.
    '''

    # ....................{ DUNDERS                        }....................
    def __getitem__(self, module_name: str) -> 'BeartypeConf':
        '''
        Return the previously instantiated beartype configuration associated
        with the module with the passed name.

        Parameters
        ----------
        module_name : str
            Fully-qualified name of the module associated with the beartype
            configuration to be returned.

        Returns
        -------
        beartype.BeartypeConf
            Beartype configuration associated with this module.

        Raises
        ------
        BeartypeClawImportConfException
            If no beartype configuration with this module has been previously
            instantiated.
        '''

        # Attempt to defer to the superclass implementation.
        try:
            return super().__getitem__(module_name)
        # If doing so fails with a low-level non-human-readable exception...
        except KeyError as exception:  # pragma: no cover
            #FIXME: Also, consider dropping the parallel
            #"BeartypeSourceFileLoader._main_module_name" attribute.
            #Does this nonsense supercede that nonsense? Probably. Which leads
            #us directly to...

            # If the module to be inspected is the "__main__" pseudo-module
            # signifying the main entry-point into the active Python process...
            if module_name == '__main__':
                # from sys import argv
                # print(f'Python arguments: "{repr(argv)}"')
                # print(f'Main module spec: "{__main__.__spec__}"')

                # Import this pseudo-module.
                #
                # Note that:
                # * Note that Python guarantees this pseudo-module to *ALWAYS*
                #   be safely importable, regardless of whether a main module
                #   actually was imported as an entry-point or not.
                # * This import *MUST* be delayed as long as feasible. In fact,
                #   this need to delay this import as long as feasible is why
                #   this import is performed here; this block is actually the
                #   last possible code path that this import can be delayed to.
                #
                # Ideally, this import would be performed earlier (e.g., in the
                # BeartypeSourceFileLoader.__init__() method defined in the
                # "beartype.claw._importlib._clawimpload" submodule) for
                # debuggability, efficiency, and idempotency; for this reason, a
                # prior implementation of the aforementioned method performed
                # this import.
                #
                # Pragmatically, doing so:
                # * Succeeded in some common edge cases, including execution of
                #   a third-party "muh_package" package invoked at the command
                #   line as "python -m muh_package", containing:
                #   * A "muh_package.__init__" submodule calling our
                #     beartype.claw.beartype_this_package() import hook.
                #   * A "muh_package.__main__" submodule.
                # * Failed in other common edge cases, including execution of a
                #   third-party "muh_package.muh_submodule" submodule invoked at
                #   the command line as "python -m muh_package.muh_submodule",
                #   containing:
                #   * A "muh_package.__init__" submodule calling our
                #     beartype.claw.beartype_this_package() import hook.
                #   * *NO* "muh_package.__main__" submodule.
                #
                # Why the discrepancy? Because CPython itself (specifically,
                # CPython's "runpy" architecture responsible for bootstrapping
                # CPython at process startup) is buggy. Due to non-trivial
                # "runpy" implementation details that are ultimately irrelevant
                # to @beartype, CPython inconsistently alters the values of
                # various critical system globals necessarily introspected by
                # the "beartype.claw" API, including erroneously reporting that:
                # * In the aforementioned "muh_package.__init__" submodule:
                #   * The "sys.argv" global is just "[-m]", thus truncating the
                #     trailing module name.
                #   * The "__main__" pseudo-module is empty and thus effectively
                #     unimportable for all intents and purposes.
                # * In the aforementioned "muh_package.__main__" and
                #   "muh_package.muh_submodule" submodules:
                #   * The "sys.argv" global is just "['muh_package']" and
                #     "['muh_package.muh_submodule']" (respectively), thus
                #     truncating the leading argument "-m".
                #   * The "__main__" pseudo-module is non-empty and thus
                #     importable for all intents and purposes.
                #
                # CPython blatantly lies about both the "sys.argv" global *AND*
                # "__main__" pseudo-module in top-level "muh_package.__init__"
                # submodules when CPython is passed the "-m" command-line
                # option. Ergo, beartype has *NO* means of introspecting either
                # object from any call in the call stack called by a top-level
                # "muh_package.__init__" submodule -- including any call to any
                # "beartype.claw" import hook. Instead, beartype *MUST* defer
                # that introspection to the last possible time... here.
                #
                # See also this relevant StackOverflow question on the topic,
                # which is nearly a decade-old as of this commit (2023 Q3) but
                # remains unresolved in even the live git version of CPython:
                #     https://stackoverflow.com/questions/42076706/sys-argv-behavior-with-python-m
                import __main__

                # Return the fully-qualified name of the actual user-defined
                # module encapsulated by the "__main__" pseudo-module.
                #
                # Note that the value of the "__main__.__name__" dunder
                # attribute is *ALWAYS* "__main__", yet another blatant lie that
                # only obfuscates the truth. Thankfully, the low-level
                # "importlib"-specific spec object publishes the fully-qualified
                # name of this actual user-defined module. I sleep now. *zzzzzz*
                return super().__getitem__(__main__.__spec__.name)  # type: ignore[union-attr]

            # Raise a high-level human-readable exception instead.
            raise BeartypeClawImportConfException(
                f'Beartype configuration associated with '
                f'module "{module_name}" hooked by '
                f'"beartype.claw" not found. '
                f'Existing beartype configurations associated with '
                f'hooked modules include:\n\t{pformat(self)}'
            ) from exception

# ....................{ CACHERS                            }....................
#FIXME: Unit test us up, please.
def cache_from_source_beartype(*args, **kwargs) -> str:
    '''
    Beartype-specific variant of the
    :func:`importlib._bootstrap_external.cache_from_source` function applying a
    beartype-specific optimization marker to that function.

    This, in turn, ensures that submodules residing in packages registered by a
    prior call to the :func:`beartype_package` function are
    compiled to files with the filetype
    ``".pyc{optimization}_{BEARTYPE_OPTIMIZATION_MARKER}"``, where
    ``{optimization}`` is the original ``optimization`` parameter passed to this
    function call.
    '''

    # Original optimization parameter passed to this function call if any *OR*
    # the empty string otherwise.
    NONBEARTYPE_OPTIMIZATION_MARKER = kwargs.get('optimization', '')

    # New optimization parameter applied by this monkey-patch of that function,
    # uniquifying that parameter with a beartype-specific suffix.
    kwargs['optimization'] = (
        f'{NONBEARTYPE_OPTIMIZATION_MARKER}{BEARTYPE_OPTIMIZATION_MARKER}')

    # Defer to the implementation of the original cache_from_source() function.
    return cache_from_source_original(*args, **kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`328`-compliant **relative module name utilities** (i.e.,
low-level callables pertaining to :pep:`328`-compliant ``"."``-prefixed relative
module names).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilModulePep328Exception
from beartype._data.typing.datatyping import (
    ListStrs,
    TypeException,
)

# ....................{ CANONICALIZERS                     }....................
def canonicalize_pep328_module_name_relative(
    # Mandatory parameters.
    module_name_relative: str,
    module_basenames_absolute: ListStrs,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilModulePep328Exception,
    exception_prefix: str = 'Module attribute ',
) -> str:
    '''
    Canonicalize (i.e., convert from relative to absolute) the passed
    :pep:`328`-compliant **relative module name** (i.e., partially-qualified
    module name prefixed by one or more ``"."`` delimiters) relative to the
    passed list of unqualified basenames comprising the fully-qualified name of
    the **parent package** (i.e., transitively containing that relative module).

    Parameters
    ----------
    module_name_relative : str
        Partially-qualified relative module name to be canonicalized into a
        fully-qualified absolute module name.
    module_basenames_absolute : list[str]
        List of the unqualified basenames comprising the fully-qualified name of
        the originating module to which this relative module is relative.
    exception_cls : type[Exception], default: _BeartypeUtilModulePep328Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilModulePep328Exception`.
    exception_prefix : str, default: "Module attribute "
        Human-readable substring prefixing raised exception messages. Defaults
        to a reasonably sensible prefix.

    Returns
    -------
    str
        Fully-qualified absolute module name canonicalized from this
        partially-qualified relative module name.

    Raises
    ------
    exception_cls
        If that module name is *not* prefixed by one or more ``"."`` delimiters
        and thus is neither relative nor :pep:`328`-compliant.
    '''
    assert isinstance(module_name_relative, str), (
        f'{repr(module_name_relative)} not string.')
    assert isinstance(module_basenames_absolute, list), (
        f'{repr(module_basenames_absolute)} not list of strings.')

    # ....................{ LOCALS                         }....................
    # The passed relative module name stripped of *ALL* prefixing "." delimiters
    # (e.g., from "..subpackage.submodule" to merely "subpackage.submodule").
    module_name_relative_stripped = module_name_relative.lstrip('.')

    # ....................{ LOCALS ~ length                }....................
    # Number of characters in this relative module name.
    module_name_relative_len = len(module_name_relative)

    # Number of "." delimiters prefixing this relative module name.
    module_name_relative_prefix_len = (
        module_name_relative_len - len(module_name_relative_stripped))

    # If this relative module name is prefixed by *NO* "." delimiters, this
    # module name is neither relative nor PEP 328-compliant. In this case...
    if not module_name_relative_prefix_len:
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not exception type.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Raise an exception.
        raise exception_cls(
            f'{exception_prefix}'
            f'PEP 328 relative module name "{module_name_relative}" '
            f'prefixed by no "." delimiters and thus '
            f'neither relative nor PEP 328-compliant.'
        )
    # Else, this relative module name is prefixed by one or more "." delimiters.

    # ....................{ LOCALS ~ slice                 }....................
    # Slice of unqualified basenames comprising the fully-qualified name of the
    # parent (sub)package to which this relative module name is relative.
    # Happily, the negative 0-based index of one after the last unqualified
    # basename of the passed list of unqualified basenames comprising the
    # fully-qualified name of the originating module such that the slice
    # "module_basenames_absolute[:module_basenames_absolute_index_last]" yields
    # the proper subset of unqualified basenames comprising the fully-qualified
    # name of the parent (sub)package to which this relative module name is
    # relative is *EXACTLY* the negation of the number of "." delimiters
    # prefixing this relative module name. Praise Guido! \o/
    module_basenames_absolute_sliced = module_basenames_absolute[
        :-module_name_relative_prefix_len]

    # Fully-qualified name of the parent (sub)package to which this relative
    # module name is relative.
    module_name_absolute_prefix = '.'.join(module_basenames_absolute_sliced)
    # print(f'module_name_absolute_prefix: {module_name_absolute_prefix}')
    # print(f'module_name_relative_stripped: {module_name_relative_stripped}')

    # Fully-qualified name of this relative module name.
    module_name = (
        f'{module_name_absolute_prefix}.{module_name_relative_stripped}'
        if module_name_relative_stripped else
        module_name_absolute_prefix
    )

    # ....................{ RETURN                         }....................
    # Return this name.
    return module_name



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
:pep:`484`-compliant **callable utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.utilfunc.pep.utilpep484func` submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                             }....................
def test_is_func_pep484_notypechecked() -> None:
    '''
    Test the
    :func:`beartype._util.func.pep.utilpep484func.is_func_pep484_notypechecked`
    tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.pep.utilpep484func import (
        is_func_pep484_notypechecked)
    from typing import no_type_check

    @no_type_check
    def now_float_above_thy_darkness() -> None:
        '''
        Arbitrary callable decorated by the
        :pep:`484`-compliant :func:`typing.no_type_check` decorator.
        '''

        pass

    def and_now_rest() -> None:
        '''
        Arbitrary callable *not* decorated by the
        :pep:`484`-compliant :func:`typing.no_type_check` decorator.
        '''

        pass

    # Assert this tester returns the expected results for these callables.
    assert is_func_pep484_notypechecked(
        now_float_above_thy_darkness) is True
    assert is_func_pep484_notypechecked(
        and_now_rest) is False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype typing callable caching** (i.e., general-purpose memoization of
function and method calls intended to be called *only* from submodules of this
subpackage) utilities.

This private submodule implements only a minimal subset of the caching
functionality implemented by the general-purpose
:mod:`beartype._util.cache.utilcachecall` submodule, from which this submodule
was originally derived. Since the latter transitively imports from the
:mod:`beartype.typing` subpackage at module scope, submodules of the
:mod:`beartype.typing` subpackage *cannot* safely import from the
:mod:`beartype._util.cache.utilcachecall` submodule at module scope. Ergo, the
existence of this submodule.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from collections.abc import Callable
from functools import wraps

# ....................{ PRIVATE ~ globals                  }....................
# Note that we intentionally avoid importing these type hint factories from
# "beartype.typing", as that would induce a circular import dependency. Instead,
# we manually import the relevant type hint factories

_Dict = dict
'''
PEP 585-compliant non-deprecated dictionary type hint factory.
'''


_SENTINEL = object()
'''
Sentinel object of arbitrary value.
'''

# ....................{ DECORATORS                         }....................
def callable_cached_minimal(func: Callable) -> Callable:
    '''
    **Memoize** (i.e., efficiently cache and return all previously returned
    values of the passed callable as well as all previously raised exceptions
    of that callable previously rather than inefficiently recalling that
    callable) the passed callable.

    Parameters
    ----------
    func : Callable
        Callable to be memoized.

    Returns
    ----------
    Callable
        Closure wrapping this callable with memoization.

    See Also
    ----------
    :func:`beartype._util.cache.utilcachecall.callable_cached`
        Further details.
    '''
    assert callable(func), f'{repr(func)} not callable.'

    # Dictionary mapping a tuple of all flattened parameters passed to each
    # prior call of the decorated callable with the value returned by that
    # call if any (i.e., if that call did *NOT* raise an exception).
    params_flat_to_return_value: _Dict[tuple, object] = {}

    # get() method of this dictionary, localized for efficiency.
    params_flat_to_return_value_get = params_flat_to_return_value.get

    # Dictionary mapping a tuple of all flattened parameters passed to each
    # prior call of the decorated callable with the exception raised by that
    # call if any (i.e., if that call raised an exception).
    params_flat_to_exception: _Dict[tuple, Exception] = {}

    # get() method of this dictionary, localized for efficiency.
    params_flat_to_exception_get = params_flat_to_exception.get

    @wraps(func)
    def _callable_cached(*args):
        f'''
        Memoized variant of the {func.__name__}() callable.

        See Also
        --------
        :func:`callable_cached`
            Further details.
        '''

        # If passed only one positional argument, minimize space consumption by
        # flattening this tuple of only that argument into that argument. Since
        # tuple items are necessarily hashable, this argument is necessarily
        # hashable as well and thus permissible as a dictionary key below.
        if len(args) == 1:
            params_flat = args[0]
        # Else, one or more positional arguments are passed. In this case,
        # reuse this tuple as is.
        else:
            params_flat = args

        # Attempt to...
        try:
            #FIXME: Optimize the params_flat_to_exception_get() case, please.
            #Since "None" is *NOT* a valid exception, we shouldn't need a
            #sentinel for safety here. Instead, this should suffice:
            #    exception = params_flat_to_exception_get(params_flat)

            #    # If this callable previously raised an exception when called with
            #    # these parameters, re-raise the same exception.
            #    if exception:
            #        raise exception

            # Exception raised by a prior call to the decorated callable when
            # passed these parameters *OR* the sentinel placeholder otherwise
            # (i.e., if this callable either has yet to be called with these
            # parameters *OR* has but failed to raise an exception).
            #
            # Note that this call raises a "TypeError" exception if any item of
            # this flattened tuple is unhashable.
            exception = params_flat_to_exception_get(params_flat, _SENTINEL)

            # If this callable previously raised an exception when called with
            # these parameters, re-raise the same exception.
            if exception is not _SENTINEL:
                raise exception  # pyright: ignore[reportGeneralTypeIssues]
            # Else, this callable either has yet to be called with these
            # parameters *OR* has but failed to raise an exception.

            # Value returned by a prior call to the decorated callable when
            # passed these parameters *OR* a sentinel placeholder otherwise
            # (i.e., if this callable has yet to be passed these parameters).
            return_value = params_flat_to_return_value_get(
                params_flat, _SENTINEL)

            # If this callable has already been called with these parameters,
            # return the value returned by that prior call.
            if return_value is not _SENTINEL:
                return return_value
            # Else, this callable has yet to be called with these parameters.

            # Attempt to...
            try:
                # Call this parameter with these parameters and cache the value
                # returned by this call to these parameters.
                return_value = params_flat_to_return_value[params_flat] = func(
                    *args)
            # If this call raises an exception...
            except Exception as exception:
                # Cache this exception to these parameters.
                params_flat_to_exception[params_flat] = exception

                # Re-raise this exception.
                raise exception
        # If one or more objects either passed to *OR* returned from this call
        # are unhashable, perform this call as is *WITHOUT* memoization. While
        # non-ideal, stability is better than raising a fatal exception.
        except TypeError:
            return func(*args)

        # Return this value.
        return return_value

    # Return this wrapper.
    return _callable_cached



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`3119`-compliant **class detectors** (i.e., callables
validating and testing various properties of arbitrary classes standardized by
:pep:`3119`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep3119Exception
from beartype.roar._roarexc import _BeartypeHintForwardRefExceptionMixin
from beartype.typing import Callable
from beartype._data.cls.datacls import TYPES_EXCEPTION_NAMESPACE
from beartype._data.typing.datatypingport import TypeIs
from beartype._data.typing.datatyping import (
    IsBuiltinOrSubclassableTypes,
    TypeException,
)
from beartype._util.cache.utilcachecall import callable_cached

# ....................{ RAISERS ~ instance                 }....................
def die_unless_object_isinstanceable(
    # Mandatory parameters.
    obj: IsBuiltinOrSubclassableTypes,

    # Optional parameters.
    is_forwardref_valid: bool = True,
    exception_cls: TypeException = BeartypeDecorHintPep3119Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception of the passed type unless the passed object is
    **isinstanceable** (i.e., valid as the second parameter to the
    :func:`isinstance` builtin).

    Specifically, this function raises an exception unless this object is
    either:

    * An **isinstanceable class** (i.e., class whose metaclass does *not* define
      an ``__instancecheck__()`` dunder method that raises a :exc:`TypeError`
      exception).
    * Tuple of one or more isinstanceable classes.
    * A :pep:`604`-compliant **new union** (i.e., objects created by expressions
      of the form ``{type1} | {type2} | ... | {typeN}``) under Python >= 3.10.
      By definition, *all* new unions are isinstanceable.

    Parameters
    ----------
    obj : IsBuiltinOrSubclassableTypes
        Object to be validated.
    is_forwardref_valid : bool, optional
        :data:`True` only if this function permits this object to be a
        **forward reference proxy** (i.e., :mod:`beartype`-specific private type
        proxying an external type that may currently be undefined). Defaults to
        :data:`True`. If this boolean is:

        * :data:`True`, this object is valid only when this object is either an
          isinstanceable classes *or* a forward reference proxy.
        * :data:`False`, this object is valid only when this object is an
          isinstanceable class. Note that forward reference proxies are
          isinstanceable classes *if and only if* the external classes they
          refer to have already been defined.
    exception_cls : TypeException, optional
        Type of exception to be raised. Defaults to
        :exc:`.BeartypeDecorHintPep3119Exception`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Raises
    ------
    BeartypeDecorHintPep3119Exception
        If this object is neither:

        * An isinstanceable class.
        * A tuple containing only isinstanceable classes.
        * A :pep:`604`-compliant new union.
    '''

    # Defer to this lower-level general-purpose raiser.
    _die_if_object_uncheckable(
        obj=obj,
        obj_pith=None,
        obj_raiser=die_unless_type_isinstanceable,
        obj_tester=isinstance,  # type: ignore[arg-type]
        is_forwardref_valid=is_forwardref_valid,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )


def die_unless_type_isinstanceable(
    # Mandatory parameters.
    cls: type,

    # Optional parameters.
    is_forwardref_valid: bool = True,
    exception_cls: TypeException = BeartypeDecorHintPep3119Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception of the passed type unless the passed object is an
    **isinstanceable class** (i.e., class whose metaclass does *not* define an
    ``__instancecheck__()`` dunder method that raises a :exc:`TypeError`
    exception).

    Classes that are *not* isinstanceable include most PEP-compliant type hints,
    notably:

    * **Generic aliases** (i.e., subscriptable classes overriding the
      ``__class_getitem__()`` class dunder method standardized by :pep:`560`
      subscripted by an arbitrary object) under Python >= 3.9, whose metaclasses
      define an ``__instancecheck__()`` dunder method to unconditionally raise
      an exception. Generic aliases include:

      * :pep:`484`-compliant **subscripted generics.**
      * :pep:`585`-compliant type hints.

    * User-defined classes whose metaclasses define an ``__instancecheck__()``
      dunder method to unconditionally raise an exception, including:

      * :pep:`544`-compliant protocols *not* decorated by the
        :func:`typing.runtime_checkable` decorator.

    Motivation
    ----------
    When a class whose metaclass defines an ``__instancecheck__()`` dunder
    method is passed as the second parameter to the :func:`isinstance` builtin,
    that builtin defers to that method rather than testing whether the first
    parameter passed to that builtin is an instance of that class. If that
    method raises an exception, that builtin raises the same exception,
    preventing callers from deciding whether arbitrary objects are instances
    of that class. For brevity, we refer to that class as "non-isinstanceable."

    Most classes are isinstanceable, because deciding whether arbitrary objects
    are instances of those classes is a core prerequisite for object-oriented
    programming. Most classes that are also PEP-compliant type hints, however,
    are *not* isinstanceable, because they're *never* intended to be
    instantiated into objects (and typically prohibit instantiation in various
    ways); they're only intended to be referenced as type hints annotating
    callables, an arguably crude form of callable markup.

    :mod:`beartype`-decorated callables typically check the types of arbitrary
    objects at runtime by passing those objects and types as the first and
    second parameters to the :func:`isinstance` builtin. If those types are
    non-isinstanceable, those type-checks will typically raise
    non-human-readable exceptions (e.g., ``"TypeError: isinstance() argument 2
    cannot be a parameterized generic"`` for :pep:`585`-compliant type hints).
    This is non-ideal both because those exceptions are non-human-readable *and*
    because those exceptions are raised at call rather than decoration time,
    where users expect the :func:`beartype.beartype` decorator to raise
    exceptions for erroneous type hints.

    Thus the existence of this function, which the :func:`beartype.beartype`
    decorator frequently calls to validate the usability of type hints that are
    classes *before* checking objects against those classes at call time.

    Caveats
    -------
    **This function considers all classes whose metaclasses define
    ``__instancecheck__()`` dunder methods that raise exceptions other than**
    :exc:`TypeError` **to be isinstanceable.** This function *only* considers
    classes whose metaclasses define ``__instancecheck__()`` dunder methods that
    raise :exc:`TypeError` exceptions to be non-isinstanceable; all other
    classes are isinstanceable.

    Ideally, this function would consider any class whose metaclass defines an
    ``__instancecheck__()`` dunder method that raises any exception (rather than
    merely a :exc:`TypeError` exception) to be non-isinstanceable.
    Pragmatically, doing so would raise false positives in common edge cases --
    and previously did so, in fact, which is why we no longer do so.

    In particular, the metaclass of the passed class may *not* necessarily be
    fully initialized at the early time that this function is called (typically,
    at :func:`beartype.beartype` decoration time). If this is the case, then
    eagerly passing that class to :func:`isinstance` is likely to raise an
    exception. For example, doing so raises an :exc:`AttributeError` when the
    ``__instancecheck__()`` dunder method defined by that metaclass references
    an external attribute that has yet to be defined:

    .. code-block:: python

       from beartype import beartype

       class MetaFoo(type):
           def __instancecheck__(cls, other):
               return g()

       class Foo(metaclass=MetaFoo):
           pass

       # @beartype transitively calls this function to validate that "Foo" is
       # isinstanceable. However, since g() has yet to be defined at this time,
       # doing so raises an "AttributeError" exception despite this logic
       # otherwise being sound.
       @beartype
       def f(x: Foo):
           pass

       def g():
           return True

    This function thus constrains itself to merely the :exc:`TypeError`
    exception, which all non-isinstanceable classes defined by the standard
    :mod:`typing` module unconditionally raise. This suggests that there is
    currently an unambiguous one-to-one mapping between non-isinstanceable
    classes and classes whose metaclass ``__instancecheck__()`` dunder methods
    raise :exc:`TypeError` exceptions. May this mapping hold true forever!

    Parameters
    ----------
    cls : object
        Object to be validated.
    is_forwardref_valid : bool, optional
        :data:`True` only if this function permits this object to be a
        **forward reference proxy** (i.e., :mod:`beartype`-specific private type
        proxying an external type that may currently be undefined). Defaults to
        :data:`True`. If this boolean is:

        * :data:`True`, this object is valid only when this object is either an
          isinstanceable classes *or* a forward reference proxy.
        * :data:`False`, this object is valid only when this object is an
          isinstanceable class. Note that forward reference proxies are
          isinstanceable classes *if and only if* the external classes they
          refer to have already been defined.
    exception_cls : TypeException, optional
        Type of exception to be raised. Defaults to
        :exc:`.BeartypeDecorHintPep3119Exception`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Raises
    ------
    BeartypeDecorHintPep3119Exception
        If this object is *not* an isinstanceable class.

    See Also
    --------
    :func:`.die_unless_type_isinstanceable`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.cls.utilclstest import die_unless_type

    # If this object is *NOT* a class, raise an exception.
    die_unless_type(
        cls=cls, exception_cls=exception_cls, exception_prefix=exception_prefix)
    # Else, this object is a class.

    # If this class is isinstanceable, silently reduce to a noop.
    #
    # Note that this merely constitutes an optimization -- albeit a
    # non-negligible optimization. The is_type_isinstanceable() tester is
    # memoized against inefficiencies, including:
    # * An inefficient implementation leveraging the EAFP (Easier to Ask for
    #   Permission than Foregiveness) principle.
    # * Worst-case behaviour in which the metaclass of this class defines an
    #   inefficient __instancecheck__() method outside our control.
    if is_type_isinstanceable(cls, is_forwardref_valid):
        return
    # Else, this class is *NOT* isinstanceable. In this case, raise a
    # human-readable exception embedding the original (typically unreadable)
    # "TypeError" exception implicitly raised by the metaclass of this class.

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with the is_type_isinstanceable() tester.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Attempt to pass this class as the second parameter to isinstance().
    try:
        isinstance(None, cls)  # type: ignore[arg-type]
    # If doing so raised *ANY* exception, this class is *NOT* isinstanceable. In
    # this case, raise a human-readable exception.
    #
    # See the docstring for further discussion.
    except Exception as exception:
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not exception class.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # If this exception ambiguously fails to indicate non-isinstanceability,
        # silently reduce to a noop.
        if _is_exception_ambiguous(exception):
            return
        # Else, this exception unambiguously indicates non-isinstanceability.
        #
        # If this exception already human-readably describes this
        # non-isinstanceability, re-raise this exception as is.
        elif _is_exception_sufficient(exception):
            raise
        # Else, this exception fails to human-readably describe this
        # non-isinstanceability.

        #FIXME: Uncomment after we uncover why doing so triggers an infinite
        #circular exception chain when "hint" is a "GenericAlias". It's clearly
        #the is_hint_pep544_protocol() call, but why? In any case, the simplest
        #workaround would just be to inline the logic of
        #is_hint_pep544_protocol() here directly. Yes, we know. *shrug*

        # # Human-readable exception message to be raised as either...
        # exception_message = (
        #     # If this class is a PEP 544-compliant protocol, a message
        #     # documenting this exact issue and how to resolve it;
        #     (
        #         f'{exception_prefix}PEP 544 protocol {hint} '
        #         f'uncheckable at runtime (i.e., '
        #         f'not decorated by @typing.runtime_checkable).'
        #     )
        #     if is_hint_pep544_protocol(hint) else
        #     # Else, a fallback message documenting this general issue.
        #     (
        #         f'{exception_prefix}type {hint} uncheckable at runtime (i.e., '
        #         f'not passable as second parameter to isinstance() '
        #         f'due to raising "{exception}" from metaclass '
        #         f'__instancecheck__() method).'
        #     )
        # )

        # Exception message to be raised.
        exception_message = (
            f'{exception_prefix}{repr(cls)} uncheckable at runtime '
            f'(i.e., not passable as second parameter to isinstance(), '
            f'due to raising "{exception.__class__.__name__}: {exception}" '
            f'from metaclass '
            f'{cls.__class__.__name__}.__instancecheck__() method).'
        )

        # Raise this exception chained onto this lower-level exception.
        raise exception_cls(exception_message) from exception

# ....................{ RAISERS ~ subclass                 }....................
def die_unless_object_issubclassable(
    # Mandatory parameters.
    obj: IsBuiltinOrSubclassableTypes,

    # Optional parameters.
    is_forwardref_valid: bool = True,
    exception_cls: TypeException = BeartypeDecorHintPep3119Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception of the passed type unless the passed object is
    **issubclassable** (i.e., valid as the second parameter to the
    :func:`issubclass` builtin).

    Specifically, this function raises an exception unless this object is
    either:

    * An **issubclassable class** (i.e., class whose metaclass does *not* define
      an ``__subclasscheck__()`` dunder method that raises a :exc:`TypeError`
      exception).
    * Tuple of one or more issubclassable classes.
    * A :pep:`604`-compliant **new union** (i.e., objects created by expressions
      of the form ``{type1} | {type2} | ... | {typeN}``) under Python >= 3.10.
      By definition, *all* new unions are issubclassable.

    Parameters
    ----------
    obj : object
        Object to be validated.
    is_forwardref_valid : bool, optional
        :data:`True` only if this function permits this object to be a
        **forward reference proxy** (i.e., :mod:`beartype`-specific private type
        proxying an external type that may currently be undefined). Defaults to
        :data:`True`. If this boolean is:

        * :data:`True`, this object is valid only when this object is either an
          isinstanceable classes *or* a forward reference proxy.
        * :data:`False`, this object is valid only when this object is an
          isinstanceable class. Note that forward reference proxies are
          isinstanceable classes *if and only if* the external classes they
          refer to have already been defined.
    exception_cls : TypeException, optional
        Type of exception to be raised. Defaults to
        :exc:`.BeartypeDecorHintPep3119Exception`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Raises
    ------
    BeartypeDecorHintPep3119Exception
        If this object is neither:

        * An issubclassable class.
        * A tuple containing only issubclassable classes.
        * A :pep:`604`-compliant new union.
    '''

    # Defer to this lower-level general-purpose raiser.
    _die_if_object_uncheckable(
        obj=obj,
        obj_pith=type,
        obj_raiser=die_unless_type_issubclassable,
        obj_tester=issubclass,  # type: ignore[arg-type]
        is_forwardref_valid=is_forwardref_valid,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )


def die_unless_type_issubclassable(
    # Mandatory parameters.
    cls: type,

    # Optional parameters.
    is_forwardref_valid: bool = True,
    exception_cls: TypeException = BeartypeDecorHintPep3119Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception of the passed type unless the passed object is an
    **issubclassable class** (i.e., class whose metaclass does *not* define a
    ``__subclasscheck__()`` dunder method that raise a :exc:`TypeError`
    exception).

    Classes that are *not* issubclassable include most PEP-compliant type
    hints, notably:

    * **Generic aliases** (i.e., subscriptable classes overriding the
      ``__class_getitem__()`` class dunder method standardized by :pep:`560`
      subscripted by an arbitrary object) under Python >= 3.9, whose
      metaclasses define an ``__subclasscheck__()`` dunder method to
      unconditionally raise an exception. Generic aliases include:

      * :pep:`484`-compliant **subscripted generics.**
      * :pep:`585`-compliant type hints.

    * User-defined classes whose metaclasses define a ``__subclasscheck__()``
      dunder method to unconditionally raise an exception, including:

      * :pep:`544`-compliant protocols *not* decorated by the
        :func:`typing.runtime_checkable` decorator.

    Motivation
    ----------
    See also the "Motivation" and "Caveats" sections of the
    :func:`.die_unless_type_isinstanceable` docstring for further discussion,
    substituting:

    * ``__instancecheck__()`` for ``__subclasscheck__()``.
    * :func:`isinstance` for :func:`issubclass`.

    Parameters
    ----------
    cls : object
        Object to be validated.
    is_forwardref_valid : bool, optional
        :data:`True` only if this function permits this object to be a
        **forward reference proxy** (i.e., :mod:`beartype`-specific private type
        proxying an external type that may currently be undefined). Defaults to
        :data:`True`. If this boolean is:

        * :data:`True`, this object is valid only when this object is either an
          isinstanceable classes *or* a forward reference proxy.
        * :data:`False`, this object is valid only when this object is an
          isinstanceable class. Note that forward reference proxies are
          isinstanceable classes *if and only if* the external classes they
          refer to have already been defined.
    exception_cls : TypeException, optional
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep3119Exception`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Raises
    ------
    BeartypeDecorHintPep3119Exception
        If this object is *not* an issubclassable class.
    '''

    # Avoid circular import dependencies.
    from beartype._util.cls.utilclstest import die_unless_type

    # If this hint is *NOT* a class, raise an exception.
    die_unless_type(
        cls=cls, exception_cls=exception_cls, exception_prefix=exception_prefix)
    # Else, this hint is a class.

    # If this class is issubclassable, silently reduce to a noop.
    #
    # Note that this merely constitutes an optimization -- albeit a
    # non-negligible optimization. The is_type_issubclassable() tester is
    # memoized against inefficiencies, including:
    # * An inefficient implementation leveraging the EAFP (Easier to Ask for
    #   Permission than Foregiveness) principle.
    # * Worst-case behaviour in which the metaclass of this class defines an
    #   inefficient __subclasscheck__() method outside our control.
    if is_type_issubclassable(cls, is_forwardref_valid):
        return
    # Else, this class is *NOT* issubclassable. In this case, raise a
    # human-readable exception embedding the original (typically unreadable)
    # "TypeError" exception implicitly raised by the metaclass of this class.

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with the is_type_issubclassable() tester.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Attempt to pass this class as the second parameter to issubclass().
    try:
        issubclass(type, cls)  # type: ignore[arg-type]
    # If doing so raised *ANY* exception, this class is *NOT* issubclassable. In
    # this case, raise a human-readable exception.
    #
    # See the die_unless_type_isinstanceable() docstring for details.
    except Exception as exception:
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not exception class.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # If this exception ambiguously fails to indicate non-issubclassability,
        # silently reduce to a noop.
        if _is_exception_ambiguous(exception):
            return
        # Else, this exception unambiguously indicates non-issubclassability.
        #
        # If this exception already human-readably describes this
        # non-issubclassability, re-raise this exception as is.
        elif _is_exception_sufficient(exception):
            raise
        # Else, this exception fails to human-readably describe this
        # non-issubclassability.

        # Exception message to be raised.
        exception_message = (
            f'{exception_prefix}{repr(cls)} uncheckable at runtime '
            f'(i.e., not passable as second parameter to issubclass(), '
            f'due to raising "{exception.__class__.__name__}: {exception}" '
            f'from metaclass '
            f'{cls.__class__.__name__}.__subclasscheck__() method).'
        )

        # Raise this exception chained onto this lower-level exception.
        raise exception_cls(exception_message) from exception

# ....................{ TESTERS ~ isinstanceable           }....................
#FIXME: Define is_object_isinstanceable() similar to as done below, please.

#FIXME: Unit test up the "is_forwardref_valid" parameter, please.
@callable_cached
def is_type_isinstanceable(
    # Mandatory parameters.
    cls: object,

    # Optional parameters.
    is_forwardref_valid: bool = True,
) -> TypeIs[type]:
    '''
    :data:`True` only if the passed object is an **isinstanceable type** (i.e.,
    class whose metaclass does *not* define an ``__instancecheck__()`` dunder
    method that raises a :exc:`TypeError` exception).

    This tester is memoized for efficiency.

    Caveats
    -------
    **This tester may return false positives in unlikely edge cases.**
    Internally, this tester tests whether this class is isinstanceable by
    detecting whether passing the :data:`None` singleton and this class to the
    :func:`isinstance` builtin raises a :exc:`TypeError` exception. If that call
    raises *no* exception, this class is probably but *not* necessarily
    isinstanceable. Since the metaclass of this class could define an
    ``__instancecheck__()`` dunder method to conditionally raise exceptions
    *except* when passed the :data:`None` singleton, there exists *no* perfect
    means of deciding whether an arbitrary class is fully isinstanceable in the
    general sense. Since most classes that are *not* isinstanceable are
    unconditionally isinstanceable (i.e., the metaclasses of those classes
    define an ``__instancecheck__()`` dunder method to unconditionally raise
    exceptions), this distinction is generally meaningless in the real world.
    This test thus generally suffices.

    Parameters
    ----------
    cls : object
        Object to be tested.
    is_forwardref_valid : bool, optional
        :data:`True` only if this function permits this object to be a
        **forward reference proxy** (i.e., :mod:`beartype`-specific private type
        proxying an external type that may currently be undefined). Defaults to
        :data:`True`. If this boolean is:

        * :data:`True`, this object is valid only when this object is either an
          isinstanceable classes *or* a forward reference proxy.
        * :data:`False`, this object is valid only when this object is an
          isinstanceable class. Note that forward reference proxies are
          isinstanceable classes *if and only if* the external classes they
          refer to have already been defined.

    Returns
    -------
    bool
        :data:`True` only if this object is an isinstanceable class.

    See Also
    --------
    :func:`.die_unless_type_isinstanceable`
        Further details.
    '''
    assert isinstance(is_forwardref_valid, bool), (
        f'{repr(is_forwardref_valid)} not bool.')

    # Avoid circular import dependencies.
    from beartype._check.forward.reference.fwdreftest import (
        is_beartype_forwardref)
    # print('Start!')

    # If this object is *NOT* a class, immediately return false.
    if not isinstance(cls, type):
        # print(f'{repr(cls)} not type!')
        return False
    # Else, this object is a class.
    #
    # If the caller allows this class to be a forward reference proxy type
    # regardless of whether the external type this proxy refers to has been
    # defined yet and this class is such a type, immediately return true.
    #
    # Note that this test is efficient and thus tested *BEFORE*
    # isinstanceability, which is less efficient.
    elif is_forwardref_valid and is_beartype_forwardref(cls):
        # print(f'{repr(cls)} is forward reference proxy!')
        return True
    # Else, either the caller prefers to disregard this distinction *OR* this
    # class is not a forward reference proxy type.
    # print('Going!')

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with die_unless_type_isinstanceable().
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Attempt to pass this class as the second parameter to the isinstance()
    # builtin to decide whether or not this class is safely usable as a
    # standard class or not.
    #
    # Note that this leverages an EAFP (i.e., "It is easier to ask forgiveness
    # than permission") approach and thus imposes a minor performance penalty,
    # but that there exists *NO* faster alternative applicable to arbitrary
    # user-defined classes, whose metaclasses may define an __instancecheck__()
    # dunder method to raise exceptions and thus prohibit being passed as the
    # second parameter to the isinstance() builtin, the primary means employed
    # by @beartype wrapper functions to check arbitrary types.
    try:
        isinstance(None, cls)  # type: ignore[arg-type]
    # If the prior function call raised *ANY* exception, return true only if
    # this exception ambiguously suggests that this class could still be
    # isinstanceable.
    except Exception as exception:
        return _is_exception_ambiguous(exception)

    # Return true, as the prior isinstance() call raised *NO* exception,
    # implying this class to probably (but *NOT* necessarily) be isinstanceable.
    return True

# ....................{ TESTERS ~ issubclassable           }....................
#FIXME: Unit test us up, please.
@callable_cached
def is_object_issubclassable(
    # Mandatory parameters.
    obj: object,

    # Optional parameters.
    is_forwardref_valid: bool = True,
) -> TypeIs[IsBuiltinOrSubclassableTypes]:
    '''
    :data:`True` only if the passed object is **issubclassable** (i.e., valid as
    the second parameter to the :func:`issubclass` builtin).

    This tester is memoized for efficiency.

    Caveats
    -------
    See also the "Caveats" sections of the
    :func:`.is_type_isinstanceable` docstring for further discussion,
    substituting:

    * ``__instancecheck__()`` for ``__subclasscheck__()``.
    * :func:`isinstance` for :func:`issubclass`.

    Parameters
    ----------
    obj : IsBuiltinOrSubclassableTypes
        Object to be tested.
    is_forwardref_valid : bool, optional
        :data:`True` only if this function permits this object to be a
        **forward reference proxy** (i.e., :mod:`beartype`-specific private type
        proxying an external type that may currently be undefined). If this
        boolean is:

        * :data:`True`, this object is valid only when this object is either:

          * An isinstanceable type.
          * A forward reference proxy (regardless of whether this proxy is
            currently resolvable to an isinstanceable type that has already been
            externally defined).

        * :data:`False`, this object is valid only when this object is either:

          * An isinstanceable type.
          * A forward reference proxy that is currently resolvable to an
            isinstanceable type that has already been externally defined.

        Defaults to :data:`True`, but probably shouldn't.

    Returns
    -------
    bool
        :data:`True` only if this object is either:

        * An issubclassable class.
        * A tuple containing only issubclassable classes.
        * A :pep:`604`-compliant new union.

    See Also
    --------
    :func:`.die_unless_type_issubclassable`
        Further details.
    '''

    # If this object is a type, defer to this lower-level type-specific tester.
    if isinstance(obj, type):
        return is_type_issubclassable(obj, is_forwardref_valid)
    # Else, this object is *NOT* a type.

    # Avoid circular import dependencies.
    from beartype._util.cls.utilclstest import is_type_or_types
    from beartype._util.hint.pep.proposal.pep484604 import is_hint_pep604

    # If this object is neither...
    if not (
        # A PEP 604-compliant new union *NOR*...
        is_hint_pep604(obj) or
        # A tuple of types...
        is_type_or_types(obj)
    # Then this object *CANNOT* be runtime-checkable. In this case, immediately
    # short-circuit by returning false.
    ):
        return False
    # Else, this object *COULD* be runtime-checkable. To decide whether this
    # object is runtime-checkable, further handling is warranted.

    #FIXME: *UGH*. DRY violation between this tester and the
    #is_type_issubclassable(). Consider refactoring out into a new private
    #_is_object_issubclassable_slow() tester, please. *sigh*

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with die_unless_type_issubclassable().
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Attempt to pass this object as the second parameter to the issubclass()
    # builtin to decide whether or not this object is safely usable as a
    # standard class or not.
    #
    # Note that this leverages an EAFP (i.e., "It is easier to ask forgiveness
    # than permission") approach and thus imposes a minor performance penalty,
    # but that there exists *NO* faster alternative applicable to arbitrary
    # user-defined classes, whose metaclasses may define a __subclasscheck__()
    # dunder method to raise exceptions and thus prohibit being passed as the
    # second parameter to the issubclass() builtin, the primary means employed
    # by @beartype wrapper functions to check arbitrary types.
    try:
        issubclass(type, obj)  # type: ignore[arg-type]

        # If the prior function call raised *NO* exception, this object is
        # probably but *NOT* necessarily issubclassable. Return true.
    # If the prior function call raised *ANY* exception, return true only if
    # this exception ambiguously suggests that this object could still be
    # issubclassable.
    except Exception as exception:
        return _is_exception_ambiguous(exception)

    # Look. Just do it. *sigh*
    return True


#FIXME: Unit test up the "is_forwardref_valid" parameter, please.
@callable_cached
def is_type_issubclassable(
    # Mandatory parameters.
    cls: object,

    # Optional parameters.
    is_forwardref_valid: bool = True,
) -> TypeIs[type]:
    '''
    :data:`True` only if the passed object is an **issubclassable class** (i.e.,
    class whose metaclass does *not* define a ``__subclasscheck__()`` dunder
    method that raises a :exc:`TypeError` exception).

    This tester is memoized for efficiency.

    Caveats
    -------
    See also the "Caveats" sections of the
    :func:`.is_type_isinstanceable` docstring for further discussion,
    substituting:

    * ``__instancecheck__()`` for ``__subclasscheck__()``.
    * :func:`isinstance` for :func:`issubclass`.

    Parameters
    ----------
    cls : object
        Object to be tested.
    is_forwardref_valid : bool, optional
        :data:`True` only if this function permits this object to be a
        **forward reference proxy** (i.e., :mod:`beartype`-specific private type
        proxying an external type that may currently be undefined). Defaults to
        :data:`True`. If this boolean is:

        * :data:`True`, this object is valid only when this object is either an
          isinstanceable classes *or* a forward reference proxy.
        * :data:`False`, this object is valid only when this object is an
          isinstanceable class. Note that forward reference proxies are
          isinstanceable classes *if and only if* the external classes they
          refer to have already been defined.

    Returns
    -------
    bool
        :data:`True` only if this object is an issubclassable class.

    See Also
    --------
    :func:`.die_unless_type_issubclassable`
        Further details.
    '''
    assert isinstance(is_forwardref_valid, bool), (
        f'{repr(is_forwardref_valid)} not bool.')

    # Avoid circular import dependencies.
    from beartype._check.forward.reference.fwdreftest import (
        is_beartype_forwardref)

    # If this object is *NOT* a class, immediately return false.
    if not isinstance(cls, type):
        return False
    # Else, this object is a class.
    #
    # If the caller allows this class to be a forward reference proxy type
    # regardless of whether the external type this proxy refers to has been
    # defined yet and this class is such a type, immediately return true.
    #
    # Note that this test is efficient and thus tested *BEFORE*
    # isinstanceability, which is less efficient.
    elif is_forwardref_valid and is_beartype_forwardref(cls):
        return True
    # Else, either the caller prefers to disregard this distinction *OR* this
    # class is not a forward reference proxy type.

    #FIXME: *UGH*. DRY violation between this tester and the
    #is_object_issubclassable(). Consider refactoring out into a new private
    #_is_object_issubclassable_slow() tester, please. *sigh*

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with die_unless_type_issubclassable().
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Attempt to pass this class as the second parameter to the issubclass()
    # builtin to decide whether or not this class is safely usable as a
    # standard class or not.
    #
    # Note that this leverages an EAFP (i.e., "It is easier to ask forgiveness
    # than permission") approach and thus imposes a minor performance penalty,
    # but that there exists *NO* faster alternative applicable to arbitrary
    # user-defined classes, whose metaclasses may define a __subclasscheck__()
    # dunder method to raise exceptions and thus prohibit being passed as the
    # second parameter to the issubclass() builtin, the primary means employed
    # by @beartype wrapper functions to check arbitrary types.
    try:
        issubclass(type, cls)  # type: ignore[arg-type]

        # If the prior function call raised *NO* exception, this class is
        # probably but *NOT* necessarily issubclassable. Return true.
    # If the prior function call raised *ANY* exception, return true only if
    # this exception ambiguously suggests that this class could still be
    # issubclassable.
    except Exception as exception:
        return _is_exception_ambiguous(exception)

    # Look. Just do it. *sigh*
    return True

# ....................{ PRIVATE ~ raisers                  }....................
def _die_if_object_uncheckable(
    obj: IsBuiltinOrSubclassableTypes,
    obj_pith: object,
    obj_raiser: Callable,
    obj_tester: Callable[[object, IsBuiltinOrSubclassableTypes], bool],
    is_forwardref_valid: bool,
    exception_cls: TypeException,
    exception_prefix: str,
) -> None:
    '''
    Raise an exception of the passed type unless the passed object is
    **runtime-checkable** (i.e., valid as the second parameter to either the
    :func:`isinstance` or :func:`issubclass` builtins) according to the passed
    object tester and raiser.

    Parameters
    ----------
    obj : object
        Object to be validated.
    obj_pith : object
        Object guaranteed to satisfy the ``obj_tester`` callable when ``obj`` is
        runtime-typecheckable (i.e., when ``obj_tester`` is called as
        ``obj_tester(obj_pith, obj)``).
    obj_raiser : Callable
        Callable raising an exception unless this object is runtime-checkable
        according to this predicate, which should be either:

        * :func:`.die_unless_type_isinstanceable`.
        * :func:`.die_unless_type_issubclassable`.
    obj_tester : Callable[[object, IsBuiltinOrSubclassableTypes], bool]
        Callable returning :data:`True` only if this object is runtime-checkable
        according to this predicate, which should be either:

        * :func:`isinstance`.
        * :func:`issubclass`.
    is_forwardref_valid : bool, optional
        :data:`True` only if this function permits this object to be a
        **forward reference proxy** (i.e., :mod:`beartype`-specific private type
        proxying an external type that may currently be undefined). Defaults to
        :data:`True`. If this boolean is:

        * :data:`True`, this object is valid only when this object is either an
          isinstanceable classes *or* a forward reference proxy.
        * :data:`False`, this object is valid only when this object is an
          isinstanceable class. Note that forward reference proxies are
          isinstanceable classes *if and only if* the external classes they
          refer to have already been defined.
    exception_cls : TypeException
        Type of exception to be raised.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message.

    Raises
    ------
    BeartypeDecorHintPep3119Exception
        If this object is *not* runtime-checkable according to the passed
        object tester and raiser.
    '''
    assert callable(obj_raiser), f'{repr(obj_raiser)} uncallable.'
    assert callable(obj_tester), f'{repr(obj_tester)} uncallable.'

    # Avoid circular import dependencies.
    from beartype._util.cls.utilclstest import die_unless_type_or_types
    from beartype._util.hint.pep.proposal.pep484604 import is_hint_pep604
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args

    # If this object is *NOT* a PEP 604-compliant new union...
    if not is_hint_pep604(obj):
        # If this object is neither a class nor tuple of classes, raise an
        # exception.
        die_unless_type_or_types(
            type_or_types=obj,
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )
        # Else, this object is either a class or tuple of classes.
    # Else, this object is a PEP 604-compliant new union.
    #
    # In any case, this object *COULD* now be runtime-checkable. To decide
    # whether this object is runtime-checkable, further handling is warranted.

    #FIXME: *UHHHH... WAT!?* This is totally busted, guys. Since this object is
    #now a PEP 604-compliant new union, this object clearly is neither a type
    #nor a tuple of types; it's a union! So, most of the code below no longer
    #makes any sense whatsoever. Actually, maybe it's just this "if
    #isinstance(obj, type):" branch that no longer makes sense? Maybe just
    #remove that branch and quietly pretend this never happened. Note that we
    #still do require the passed "obj_raiser" parameter, however. I sigh. *sigh*

    # If this object is a class...
    if isinstance(obj, type):
        # If this class is *NOT* runtime-checkable, raise an exception.
        obj_raiser(
            cls=obj,
            is_forwardref_valid=is_forwardref_valid,
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )
        # Else, this class is runtime-checkable.
    # Else, this object *MUST* (by process of elimination and validation above)
    # be either a tuple of classes *OR* a PEP 604-compliant new union. In either
    # case...
    else:
        # Attempt to pass this object as the second parameter to isinstance().
        try:
            obj_tester(obj_pith, obj)  # type: ignore[arg-type]
        # If doing so raises *ANY* exception, this object is *NOT*
        # runtime-checkable. In this case, raise a human-readable exception.
        #
        # See the die_unless_type_isinstanceable() docstring for details.
        except Exception as exception:
            assert isinstance(exception_cls, type), (
                f'{repr(exception_cls)} not exception class.')
            assert isinstance(exception_prefix, str), (
                f'{repr(exception_prefix)} not string.')

            # If this exception ambiguously fails to indicate
            # non-isinstanceability, silently reduce to a noop.
            if _is_exception_ambiguous(exception):
                return
            # Else, this exception unambiguously indicates
            # non-isinstanceability.

            # Tuple of all items of this iterable object.
            obj_items: tuple = None  # type: ignore[assignment]

            # Human-readable label describing this object in this exception
            # message.
            obj_label: str = None  # type: ignore[assignment]

            # Human-readable label describing the first non-runtime-checkable
            # item of this object in this exception message.
            obj_item_label: str = None  # type: ignore[assignment]

            # If this object is a tuple, define these locals accordingly.
            if isinstance(obj, tuple):
                obj_items = obj
                obj_label = 'tuple union'
                obj_item_label = 'tuple union item'
            # Else, this object is a new union. Define these locals accordingly.
            else:
                obj_items = get_hint_pep_args(obj)
                obj_label = 'PEP 604 new union'
                obj_item_label = 'new union child type'

            # Exception message to be raised.
            exception_message = (
                f'{exception_prefix} {obj_label} {repr(obj)} '
                f'uncheckable at runtime'
            )

            # For the 0-based index of each tuple class and that class...
            for cls_index, cls in enumerate(obj_items):
                # If this class is *NOT* runtime-checkable, raise an exception.
                obj_raiser(
                    cls=cls,
                    exception_cls=exception_cls,
                    exception_prefix=(
                        f'{exception_message}, as '
                        f'{obj_item_label} {cls_index} '
                    ),
                )
                # Else, this class is runtime-checkable. Continue to the next.

            # Raise this exception chained onto this lower-level exception.
            # Although this should *NEVER* happen (as we should have already
            # raised an exception above), we nonetheless do so for safety.
            raise exception_cls(f'{exception_message}.') from exception

# ....................{ PRIVATE ~ testers                  }....................
#FIXME: Unit test us up, please.
def _is_exception_ambiguous(exception: Exception) -> bool:
    '''
    :data:`True` only if the passed exception is **ambiguous** (i.e., an
    instance of the standard :exc:`AttributeError` or :exc:`NameError` types).

    Equivalently, this tester returns :data:`True` only if it cannot be decided
    whether this exception unambiguously signifies the failure of an associated
    class to be isinstanceable or issubclassable when the metaclass of this
    class defines the ``__instancecheck__()`` or ``__subclasscheck__()`` dunder
    method raising this exception. Why? Because this metaclass may *not*
    necessarily be fully initialized at the early time that this tester is
    called (typically, at :func:`beartype.beartype` decoration time). When this
    is the case, eagerly passing this class to the :func:`isinstance` or
    :func:`issubclass` builtins is likely to raise an :exc:`AttributeError`
    after referencing an undefined external attribute: e.g.,

    .. code-block:: python

       from beartype import beartype

       class MetaFoo(type):
           def __instancecheck__(cls, other):
               return g()

       class Foo(metaclass=MetaFoo):
           pass

       # @beartype transitively calls this function to validate that "Foo" is
       # isinstanceable. However, since g() has yet to be defined at this time,
       # doing so raises an "AttributeError" exception despite this logic
       # otherwise being sound.
       @beartype
       def f(x: Foo):
           pass

       def g():
           return True

    Parameters
    ----------
    exception : Exception
        Exception previously raised by a recent call to either the
        ``__instancecheck__()`` or ``__subclasscheck__()`` dunder method defined
        by the metaclass of the class passed as the second parameter to the
        :func:`isinstance` or :func:`issubclass` builtins.

    Returns
    -------
    bool
        :data:`True` only if this exception is ambiguous.
    '''

    # Faster than one-liner speed!
    return isinstance(exception, TYPES_EXCEPTION_NAMESPACE)


def _is_exception_sufficient(exception: Exception) -> bool:
    '''
    :data:`True` only if the passed exception is **sufficient** (i.e., an
    instance of the :mod:`beartype`-specific
    :exc:`._BeartypeHintForwardRefExceptionMixin` type).

    Equivalently, this tester returns :data:`True` only if this exception's
    message already suffices to human-readably describe the failure of an
    associated class to be isinstanceable or issubclassable when the metaclass
    of this class defines the ``__instancecheck__()`` or ``__subclasscheck__()``
    dunder method raising this exception. Currently, this includes *all*
    exceptions pertaining to :mod:`beartype`-specific forward references. Why?
    Because these exceptions already convey the issue. Wrapping these exceptions
    in additional exceptions only obfuscates the underlying issue: e.g.,

    .. code-block:: python

       # This original exception...
       BeartypeCallHintForwardRefException: Forward reference "OfPearl"
       unimportable from module
       "beartype_test.a00_unit.data.pep.pep563.data_pep563_resolve".

       # ...is considerably more useful to end users than this insane wrapper.
       beartype.roar.BeartypeDecorHintNonpepException: Die_if_unbearable()
       <forwardref OfPearl(__name_beartype__='OfPearl',
       __scope_name_beartype__='beartype_test.a00_unit.data.pep.pep563.data_pep563_resolve')>
       uncheckable at runtime (i.e., not passable as second parameter to
       isinstance(), due to raising "BeartypeCallHintForwardRefException:
       Forward reference "OfPearl" unimportable from module
       "beartype_test.a00_unit.data.pep.pep563.data_pep563_resolve"." from
       metaclass BeartypeForwardRefMeta.__instancecheck__() method).

    Parameters
    ----------
    exception : Exception
        Exception previously raised by a recent call to either the
        ``__instancecheck__()`` or ``__subclasscheck__()`` dunder method defined
        by the metaclass of the class passed as the second parameter to the
        :func:`isinstance` or :func:`issubclass` builtins.

    Returns
    -------
    bool
        :data:`True` only if this exception is sufficient.
    '''

    # Fear the one-liner reaper!
    return isinstance(exception, _BeartypeHintForwardRefExceptionMixin)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Testing-specific **typing attribute utilities** (i.e., functions introspecting
attributes with arbitrary names dynamically imported from typing modules).
'''

# ....................{ TODO                               }....................
#FIXME: Consider excising this submodule. Ideally, *ALL* functionality defined
#by this submodule should instead reside in the central
#"beartype._util.api.standard.utiltyping" submodule, which significantly overlaps
#with this submodule in harmful way. We appear to have duplicated work between
#these two submodules unknowingly, which is simply horrid. Gah! I cry at night.

# ....................{ IMPORTS                            }....................
from typing import Any

# ....................{ IMPORTERS                          }....................
def import_typing_attr_or_none_safe(typing_attr_basename: str) -> Any:
    '''
    Dynamically import and return the **typing attribute** (i.e., object
    declared at module scope by either the :mod:`typing` or
    :mod:`typing_extensions` modules) with the passed unqualified name if
    importable from one or more of these modules *or* ``None`` otherwise
    otherwise (i.e., if this attribute is *not* importable from these modules).

    Caveats
    -------
    **This higher-level wrapper should typically be called in lieu of the
    lower-level**
    :func:`beartype._util.api.standard.utiltyping.import_typing_attr_or_none`
    **function.** Unlike the latter, this wrapper imports from the third-party
    :mod:`typing_extensions` module *only* if the version of that module is
    sufficiently new and thus satisfies test-time requirements.

    Parameters
    ----------
    typing_attr_basename : str
        Unqualified name of the attribute to be imported from a typing module.

    Returns
    -------
    object
        Attribute with this name dynamically imported from a typing module.

    Raises
    ------
    beartype.roar._roarexc._BeartypeUtilModuleException
        If this name is syntactically invalid.

    Warns
    -----
    BeartypeModuleUnimportableWarning
        If any of these modules raise module-scoped exceptions at importation
        time. That said, the :mod:`typing` and :mod:`typing_extensions` modules
        are scrupulously tested and thus unlikely to raise such exceptions.
    '''

    # Defer test-specific imports.
    from beartype._util.api.standard.utiltyping import import_typing_attr_or_none
    from beartype._util.module.utilmodimport import import_module_attr_or_none
    from beartype_test._util.module.pytmodtest import (
        is_package_typing_extensions)
    # print(f'is_package_typing_extensions: {is_package_typing_extensions()}')

    # If a reasonably recent version of the third-party "typing_extensions"
    # package is importable under the active Python interpreter, defer to this
    # higher-level importer possibly returning an attribute from that package.
    if is_package_typing_extensions():
        return import_typing_attr_or_none(typing_attr_basename)
    # Else, either "typing_extensions" is unimportable or only an obsolete
    # version of "typing_extensions" is importable; in either case, avoid
    # possibly returning a possibly broken attribute from that package by
    # importing only from the official "typing" module.

    # Return the typing attribute imported from the official "typing" module if
    # defined by this version of Python *OR* "None" otherwise.
    return import_module_attr_or_none(f'typing.{typing_attr_basename}')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **beartype import hook decorator-hostile decorator subpackage
initialization submodule** (i.e., data module mimicking real-world usage of
various :func:`beartype.claw` import hooks on callables decorated by chains of
one or more **decorator-hostile decorators** (i.e., decorators hostile to other
decorators by prohibiting other decorators from being applied after they
themselves are applied in such a chain)).
'''

# ....................{ IMPORTS                            }....................
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.decor_hostile import (
    data_claw_decor_hostile_func,
    data_claw_decor_hostile_type,
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-compliant type hint tester** (i.e., callable validating an
arbitrary object to be a PEP-compliant type hint) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import (
    BeartypeDecorHintPepException,
    BeartypeDecorHintPepUnsupportedException,
    BeartypeDecorHintPep484Exception,
)
from beartype.typing import (
    NoReturn,
)
from beartype._data.typing.datatypingport import (
    Hint,
    TypeIs,
)
from beartype._data.typing.datatyping import TypeException
from beartype._data.hint.sign.datahintsignset import (
    HINT_SIGNS_SUPPORTED,
    HINT_SIGNS_TYPE_MIMIC,
)
from beartype._data.api.standard.datatyping import TYPING_MODULE_NAMES
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.module.utilmodget import get_object_module_name_or_none
from beartype._util.utilobject import get_object_type_unless_type

# ....................{ EXCEPTIONS                         }....................
def die_if_hint_pep(
    # Mandatory parameters.
    hint: object,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPepException,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception of the passed type if the passed object is a
    **PEP-compliant type hint** (i.e., :mod:`beartype`-agnostic annotation
    compliant with annotation-centric PEPs).

    This validator is effectively (but technically *not*) memoized. See the
    :func:`beartype._util.hint.utilhinttest.die_unless_hint` validator.

    Parameters
    ----------
    hint : object
        Object to be validated.
    exception_cls : Type[Exception], optional
        Type of the exception to be raised by this function. Defaults to
        :exc:`.BeartypeDecorHintPepException`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Raises
    ------
    exception_cls
        If this object is a PEP-compliant type hint.
    '''

    # If this hint is PEP-compliant...
    if is_hint_pep(hint):
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not type.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Raise an exception of this class.
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} is PEP-compliant '
            f'(e.g., rather than isinstanceable class).'
        )


def die_unless_hint_pep(
    # Mandatory parameters.
    hint: object,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPepException,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception unless the passed object is a **PEP-compliant type
    hint** (i.e., :mod:`beartype`-agnostic annotation compliant with
    annotation-centric PEPs).

    This validator is effectively (but technically *not*) memoized. See also the
    :func:`beartype._util.hint.utilhinttest.die_unless_hint` validator.

    Parameters
    ----------
    hint : object
        Object to be validated.
    exception_cls : Type[Exception], optional
        Type of the exception to be raised by this function. Defaults to
        :class:`.BeartypeDecorHintPepException`.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If this object is *not* a PEP-compliant type hint.
    '''

    # If this hint is *NOT* PEP-compliant, raise an exception.
    if not is_hint_pep(hint):
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not type.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} not PEP-compliant.')

# ....................{ EXCEPTIONS ~ supported             }....................
def die_if_hint_pep_unsupported(
    # Mandatory parameters.
    hint: object,

    # Optional parameters.
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception if the passed object is a **PEP-compliant unsupported
    type hint** (i.e., :mod:`beartype`-agnostic annotation compliant with
    annotation-centric PEPs currently *not* supported by the
    :func:`beartype.beartype` decorator).

    This validator is effectively (but technically *not*) memoized. See the
    :func:`beartype._util.hint.utilhinttest.die_unless_hint` validator.

    Caveats
    -------
    **This validator only shallowly validates this object.** If this object is a
    subscripted PEP-compliant type hint (e.g., ``Union[str, List[int]]``), this
    validator ignores all subscripted arguments (e.g., ``List[int]``) on this
    hint and may thus return false positives for hints that are directly
    supported but whose subscripted arguments are not. To deeply validate this
    object, iteratively call this validator during a recursive traversal (such
    as a breadth-first search) over each subscripted argument of this object.

    Parameters
    ----------
    hint : object
        Object to be validated.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    BeartypeDecorHintPepException
        If this object is *not* a PEP-compliant type hint.
    BeartypeDecorHintPepUnsupportedException
        If this object is a PEP-compliant type hint but is currently
        unsupported by the :func:`beartype.beartype` decorator.
    BeartypeDecorHintPep484Exception
        If this object is the PEP-compliant :attr:`typing.NoReturn` type hint,
        which is contextually valid in only a single use case and thus
        supported externally by the :mod:`beartype._decor._nontype._wrap.wrapmain`
        submodule rather than with general-purpose automation.
    '''

    # If this object is a supported PEP-compliant type hint, reduce to a noop.
    #
    # Note that this memoized call is intentionally passed positional rather
    # than keyword parameters to maximize efficiency.
    if is_hint_pep_supported(hint):
        return
    # Else, this object is *NOT* a supported PEP-compliant type hint. In this
    # case, subsequent logic raises an exception specific to the passed
    # parameters.

    # If this hint is *NOT* PEP-compliant, raise a more readable exception.
    die_unless_hint_pep(hint=hint, exception_prefix=exception_prefix)
    assert isinstance(exception_prefix, str), (
        f'{repr(exception_prefix)} not string.')
    # Else, this hint is PEP-compliant.

    #FIXME: Pretty weird, honestly. This would almost certainly better be
    #raised in a reducer testing whether "arg_name == ARG_NAME_RETURN".
    # If this is the PEP 484-compliant "typing.NoReturn" type hint permitted
    # *ONLY* as a return annotation, raise an exception specific to this hint.
    if hint is NoReturn:
        raise BeartypeDecorHintPep484Exception(
            f'{exception_prefix}PEP 484 type hint "{repr(hint)}" '
            f'invalid in this type hint context (i.e., '
            f'"{repr(hint)}" valid only as non-nested return annotation).'
        )
    # Else, this is any PEP-compliant type hint other than "typing.NoReturn".

    # In this case, raise a general-purpose exception.
    #
    # Note that, by definition, the sign uniquely identifying this hint *SHOULD*
    # be in the "HINT_SIGNS_SUPPORTED" set. Regardless of whether it is or not,
    # we raise a similar exception in either case. Ergo, there is *NO* practical
    # benefit to validating that expectation here.
    raise BeartypeDecorHintPepUnsupportedException(
        f'{exception_prefix}type hint {repr(hint)} '
        f'currently unsupported by @beartype.'
    )

# ....................{ WARNINGS                           }....................
#FIXME: Unit test us up.
#FIXME: Actually use us in place of die_if_hint_pep_unsupported().
#FIXME: Actually, it's unclear whether we still require or desire this. See
#"_pephint" commentary for further details.
# def warn_if_hint_pep_unsupported(
#     # Mandatory parameters.
#     hint: object,
#
#     # Optional parameters.
#     exception_prefix: str = 'Annotated',
# ) -> bool:
#     '''
#     Return ``True`` and emit a non-fatal warning only if the passed object is a
#     **PEP-compliant unsupported type hint** (i.e., :mod:`beartype`-agnostic
#     annotation compliant with annotation-centric PEPs currently *not* supported
#     by the :func:`beartype.beartype` decorator).
#
#     This validator is effectively (but technically *not*) memoized. See the
#     :func:`beartype._util.hint.utilhinttest.die_unless_hint` validator.
#
#     Parameters
#     ----------
#     hint : object
#         Object to be validated.
#     exception_prefix : Optional[str]
#         Human-readable label prefixing this object's representation in the
#         warning message emitted by this function. Defaults to the empty string.
#
#     Returns
#     ----------
#     bool
#         ``True`` only if this PEP-compliant type hint is currently supported by
#         that decorator.
#
#     Raises
#     ----------
#     BeartypeDecorHintPepException
#         If this object is *not* a PEP-compliant type hint.
#
#     Warnings
#     ----------
#     BeartypeDecorHintPepUnsupportedWarning
#         If this object is a PEP-compliant type hint currently unsupported by
#         that decorator.
#     '''
#
#     # True only if this object is a supported PEP-compliant type hint.
#     #
#     # Note that this memoized call is intentionally passed positional rather
#     # than keyword parameters to maximize efficiency.
#     is_hint_pep_supported_test = is_hint_pep_supported(hint)
#
#     # If this object is an unsupported PEP-compliant type hint...
#     if not is_hint_pep_supported_test:
#         assert isinstance(exception_prefix, str), f'{repr(exception_prefix)} not string.'
#
#         # If this hint is *NOT* PEP-compliant, raise an exception.
#         die_unless_hint_pep(hint=hint, exception_prefix=exception_prefix)
#
#         # Else, this hint is PEP-compliant. In this case, emit a warning.
#         warn(
#             (
#                 f'{exception_prefix}PEP type hint {repr(hint)} '
#                 f'currently unsupported by @beartype.'
#             ),
#             BeartypeDecorHintPepUnsupportedWarning
#         )
#
#     # Return true only if this object is a supported PEP-compliant type hint.
#     return is_hint_pep_supported_test

# ....................{ TESTERS                            }....................
def is_hint_pep(hint: object) -> TypeIs[Hint]:
    '''
    :data:`True` only if the passed object is a **PEP-compliant type hint**
    (i.e., object complying with one or more :mod:`typing`-centric Python
    Enhancement Proposals (PEPs)).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Motivation
    ----------
    Standard Python types allow callers to test for compliance with protocols,
    interfaces, and abstract base classes by calling either the
    :func:`isinstance` or :func:`issubclass` builtins. This is the
    well-established Pythonic standard for deciding conformance to an API.

    Insanely, :pep:`484` *and* the :mod:`typing` module implementing :pep:`484`
    reject community standards by explicitly preventing callers from calling
    either the :func:`isinstance` or :func:`issubclass` builtins on most but
    *not* all :pep:`484` objects and types. Moreover, neither :pep:`484` nor
    :mod:`typing` implement public APIs for testing whether arbitrary objects
    comply with :pep:`484` or :mod:`typing`.

    Thus this function, which "fills in the gaps" by implementing this
    laughably critical oversight.

    Parameters
    ----------
    hint : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a PEP-compliant type hint.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

    # Sign uniquely identifying this hint if this hint is PEP-compliant *OR*
    # "None" otherwise (i.e., if this hint is *NOT* PEP-compliant).
    hint_sign = get_hint_pep_sign_or_none(hint)  # pyright: ignore
    # print(f'hint: {repr(hint)}; sign: {repr(hint_sign)}')

    # Return true *ONLY* if this hint is uniquely identified by a sign and thus
    # PEP-compliant.
    return hint_sign is not None


#FIXME: Currently unused but preserved for posterity. *shrug*
# def is_hint_pep_deprecated(hint: object) -> bool:
#     '''
#     :data:`True` only if the passed PEP-compliant type hint is **deprecated**
#     (i.e., obsoleted by an equivalent PEP-compliant type hint standardized by a
#     more recently released PEP).
#
#     This tester is intentionally *not* memoized (e.g., by the
#     ``callable_cached`` decorator), as this tester is currently *only* called at
#     test time from our test suite.
#
#     Parameters
#     ----------
#     hint : object
#         PEP-compliant type hint to be inspected.
#
#     Returns
#     -------
#     bool
#         :data:`True` only if this PEP-compliant type hint is deprecated.
#     '''
#
#     # Avoid circular import dependencies.
#     from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign
#
#     # Sign uniquely identifying this hint.
#     hint_sign = get_hint_pep_sign(hint)
#
#     # Return true only if either...
#     return (
#         # This sign is that of an unconditionally deprecated type hint *OR*...
#         hint_sign in HINT_SIGNS_DEPRECATED or
#         # This is a PEP 484-compliant type hint (e.g., "typing.List[str]")
#         # conditionally deprecated by an equivalent PEP 585-compliant type hint
#         # (e.g., "list[str]") under Python >= 3.9.
#         #
#         # Note that, in this case, the sign of this hint does *NOT* convey
#         # enough metadata to ascertain whether this hint is deprecated. Ergo, a
#         # non-trivial tester dedicated to this discernment is required: e.g.,
#         # * "list[str]" has the sign "HintSignList" but is *NOT* deprecated.
#         # * "typing.List[str]" has the sign "HintSignList" but is deprecated.
#         is_hint_pep484_deprecated(hint)
#     )


#FIXME: Unit test us up, please.
def is_hint_pep_subbed(hint: object) -> bool:
    '''
    :data:`True` only if the passed PEP-compliant type hint is **subscripted**
    (i.e., indexed by one or more child type hints).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`beartype._util.cache.utilcachecall.callable_cached` decorator), as
    the implementation trivially reduces to an efficient one-liner.

    Parameters
    ----------
    hint : object
        PEP-compliant type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this PEP-compliant type hint is subscripted.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args

    # Return true only if the tuple of all child type hints subscripting this
    # hint contains one or more child type hints.
    return bool(get_hint_pep_args(hint))


@callable_cached
def is_hint_pep_supported(hint: object) -> bool:
    '''
    :data:`True` only if the passed object is a **PEP-compliant supported type
    hint** (i.e., :mod:`beartype`-agnostic annotation compliant with
    annotation-centric PEPs currently supported by the
    :func:`beartype.beartype` decorator).

    This tester is memoized for efficiency.

    Caveats
    -------
    **This tester only shallowly inspects this object.** If this object is a
    subscripted PEP-compliant type hint (e.g., ``Union[str, List[int]]``), this
    tester ignores all subscripted arguments (e.g., ``List[int]``) on this hint
    and may thus return false positives for hints that are directly supported
    but whose subscripted arguments are not.

    To deeply inspect this object, iteratively call this tester during a
    recursive traversal over each subscripted argument of this object.

    Parameters
    ----------
    hint : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a supported PEP-compliant type hint.
    '''

    # If this hint is *NOT* PEP-compliant, immediately return false.
    if not is_hint_pep(hint):
        return False
    # Else, this hint is PEP-compliant.

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign

    # Sign uniquely identifying this hint.
    hint_sign = get_hint_pep_sign(hint)

    # Return true only if this sign is supported.
    return hint_sign in HINT_SIGNS_SUPPORTED

# ....................{ TESTERS ~ typing                   }....................
#FIXME: Replace all hardcoded "'typing" strings throughout the codebase with
#access of "TYPING_MODULE_NAMES" instead. We only see one remaining in:
#* beartype._util.hint.pep.proposal.pep484.pep484.py
#
#Thankfully, nobody really cares about generalizing that one edge case to
#"testing_extensions", so it's mostly fine for various definitions of fine.
def is_hint_pep_typing(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed type hint is an attribute of a **typing
    module** (i.e., module officially declaring attributes usable for creating
    PEP-compliant type hints accepted by both static and runtime type checkers).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this type hint is an attribute of a typing module.
    '''
    # print(f'is_hint_pep_typing({repr(hint)}')

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

    # Return true only if this hint is either...
    return (
        # Any PEP-compliant type hint defined by a typing module (except those
        # maliciously masquerading as another type entirely) *OR*...
        get_object_module_name_or_none(hint) in TYPING_MODULE_NAMES or
        # Any PEP-compliant type hint defined by a typing module maliciously
        # masquerading as another type entirely.
        get_hint_pep_sign_or_none(hint) in HINT_SIGNS_TYPE_MIMIC  # pyright: ignore
    )


def is_hint_pep_type_typing(hint: Hint) -> bool:
    '''
    :data:`True` only if either the passed type hint is defined by a **typing
    module** (i.e., module officially declaring attributes usable for creating
    PEP-compliant type hints accepted by both static and runtime type checkers)
    if this hint is a class *or* the class of this hint is defined by a typing
    module otherwise (i.e., if this object is *not* a class).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if either:

        * If this hint is a class, this class is defined by a typing module.
        * Else, the class of this hint is defined by a typing module.
    '''

    # This hint if this hint is a class *OR* this hint's class otherwise.
    hint_type = get_object_type_unless_type(hint)
    # print(f'pep_type_typing({repr(hint)}): {get_object_module_name(hint_type)}')

    # Return true only if this type is defined by a typing module.
    #
    # Note that this implementation could probably be reduced to the
    # leading portion of the body of the get_hint_pep_sign_or_none()
    # function testing this object's representation. While certainly more
    # compact and convenient than the current approach, that refactored
    # approach would also be considerably more fragile, failure-prone, and
    # subject to whimsical "improvements" in the already overly hostile
    # "typing" API. Why? Because the get_hint_pep_sign_or_none() function:
    # * Parses the machine-readable string returned by the __repr__()
    #   dunder method of "typing" types. Since that string is *NOT*
    #   standardized by PEP 484 or any other PEP, "typing" authors remain
    #   free to violate this pseudo-standard in any manner and at any time
    #   of their choosing.
    # * Suffers common edge cases for "typing" types whose __repr__()
    #   dunder methods fail to comply with the non-standard implemented by
    #   their sibling types. This includes the common "TypeVar" type.
    # * Calls this tester function to decide whether the passed object is a
    #   PEP-compliant type hint or not before subjecting that object to
    #   further introspection, which would clearly complicate implementing
    #   this tester function in terms of that getter function.
    #
    # In contrast, the current approach only tests the standard
    # "__module__" dunder attribute and is thus significantly more robust
    # against whimsical destruction by "typing" authors. Note that there
    # might exist an alternate means of deciding this boolean, documented
    # here merely for completeness:
    #     try:
    #         isinstance(obj, object)
    #         return False
    #     except TypeError as type_error:
    #         return str(type_error).endswith(
    #             'cannot be used with isinstance()')
    #
    # The above effectively implements an Aikido throw by using the fact
    # that "typing" types prohibit isinstance() calls against those types.
    # While clever (and deliciously obnoxious), the above logic:
    # * Requires catching exceptions in the common case and is thus *MUCH*
    #   less efficient than the preferable approach implemented here.
    # * Assumes that *ALL* "typing" types prohibit such calls. Sadly, only
    #   a proper subset of these types prohibit such calls.
    # * Assumes that those "typing" types that do prohibit such calls raise
    #   exceptions with reliable messages across *ALL* Python versions.
    #
    # In short, there is no general-purpose clever solution. *sigh*
    return hint_type.__module__ in TYPING_MODULE_NAMES



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **forward reference metaclasses** (i.e., low-level metaclasses of
classes deferring the resolution of a stringified type hint referencing an
attribute that has yet to be defined and annotating a class or callable
decorated by the :func:`beartype.beartype` decorator).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintForwardRefException
from beartype.typing import Dict
from beartype._data.typing.datatyping import BeartypeForwardRef
from beartype._util.cls.pep.clspep3119 import (
    die_unless_object_isinstanceable)
# from beartype._util.func.utilfuncframe import is_frame_caller_beartype
from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
    get_hint_pep484585_generic_type)
from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
    is_hint_pep484585_generic)
from beartype._util.module.utilmodimport import import_module_attr
from beartype._util.text.utiltextidentifier import is_dunder

# ....................{ METACLASSES                        }....................
class BeartypeForwardRefMeta(type):
    '''
    **Forward reference metaclass** (i.e., metaclass of the
    :class:`.BeartypeForwardRefABC` superclass deferring the resolution of a
    stringified type hint referencing an attribute that has yet to be defined
    and annotating a class or callable decorated by the
    :func:`beartype.beartype` decorator).

    This metaclass memoizes each **forward reference** (i.e.,
    :class:`.BeartypeForwardRefABC` instance) according to the fully-qualified
    name of the attribute referenced by that forward reference. Doing so ensures
    that only the first :class:`.BeartypeForwardRefABC` instance referring to a
    unique attribute is required to dynamically resolve that attribute at
    runtime; all subsequent :class:`.BeartypeForwardRefABC` instances referring
    to the same attribute transparently reuse the attribute previously resolved
    by the first such instance, effectively reducing the time cost of resolving
    forward references to a constant-time operation with negligible constants.

    This metaclass dynamically and efficiently resolves each forward reference
    in a just-in-time (JIT) manner on the first :func:`isinstance` call whose
    second argument is that forward reference. Forward references *never* passed
    to the :func:`isinstance` builtin are *never* resolved, which is good.
    '''

    # ....................{ DUNDERS                        }....................
    #FIXME: That's great, but still insufficient. Additionally:
    #* If the caller resides in a "beartype."-prefixed submodule, do what we
    #  currently do.
    #* Else, immediately resolve the referent by accessing "__type_beartype__"
    #  and then (as above) proxy the __getattr__() of this referent by calling
    #  getattr() against this referent.
    def __getattr__(  # type: ignore[misc]
        cls: BeartypeForwardRef, hint_name: str) -> BeartypeForwardRef:
        '''
        **Fully-qualified forward reference subclass** (i.e.,
        :class:`.BeartypeForwardRefABC` subclass whose metaclass is this
        metaclass and whose :attr:`.BeartypeForwardRefABC.__name_beartype__`
        class variable is the fully-qualified name of an external class).

        This dunder method creates and returns a new forward reference subclass
        referring to an external class whose name is concatenated from (in
        order):

        #. The fully-qualified name of the external package or module referred
           to by the passed forward reference subclass.
        #. The passed unqualified basename, presumably referring to a
           subpackage, submodule, or class of that external package or module.

        Design
        ------
        The syntactic implementation of this dunder method is largely trivial.
        The semantic justification for this implementation is, however, anything
        but. Indeed, justifying this implementation warrants a full-length
        dissertation on runtime resolution of forward references. This is that
        dissertation.

        Broadly speaking, there are two use cases for which CPython implicitly
        invokes this dunder method: two use cases whose intentions and
        requirements are so at odds with one another that seamlessly satisfying
        both is an exercise in code torture.

        The first use case is the intended (and also most common) use case:
        **absolute forward reference resolution deferral.** Given a
        :pep:`484`-compliant stringified absolute forward reference to a
        subscripted generic that has yet to be defined (e.g.,
        ``"some_package.some_submodule.SomeType[T]"``), how *exactly* does
        :mod:`beartype` resolve that subscripted generic in a manner consistent
        with efficient runtime type-checking? Is such resolution even feasible?
        The answers, of course, are: "Carefully." and "Yuppers."

        One brute-force approach to resolving stringified forward references
        containing arbitrarily complex Python expressions at runtime would be to
        parse those references through a Python-specific Parser Expression
        Grammar (PEG). Although technically feasible, embedding a full-blown
        Python parser within :mod:`beartype` would be so fragile and inefficient
        as to be effectively infeasible. Consequently, :mod:`beartype` does
        *not* do that. Instead, :mod:`beartype` is clever.

        The clever approach is to charm Python itself into parsing those
        references. After all, Python clearly knows how to parse Python.
        :mod:`beartype` simply needs to transform those references into some
        format readily digestible by Python's builtin Python parser. Our
        solution? The :func:`eval` builtin coupled with our non-standard
        :class:`beartype._check.forward.fwdscope.BeartypeForwardScope`
        dictionary subclass, which overrides the ``__missing__`` dunder method
        explicitly called by the superclass :meth:`dict.__getitem__` method
        implicitly called on each ``[``- and ``]``-delimited attempt to access a
        forward reference whose type has yet to be resolved by mapping the name
        of that reference to an actual **forward reference proxy** (i.e.,
        instance of this metaclass). The :func:`eval` builtin then implicitly
        instantiates one forward reference proxy encapsulating each undefined
        top-level attribute inside the passed absolute forward reference. Given
        ``"some_package.some_submodule.SomeType[T]"``, :func:`eval` then first
        instantiates one forward reference proxy encapsulating the undefined
        top-level attribute ``"some_package"``. Clearly, a forward reference
        proxy for ``"some_package"`` does *not* suffice to proxy the entire
        ``"some_package.some_submodule.SomeType[T]"`` forward reference.

        Cue this dunder method. :func:`eval` then attempts to access the
        ``"some_submodule"`` attribute of the forward reference proxy for
        ``"some_package"``. Doing so implicitly invokes this dunder method,
        which then instantiates another forward reference proxy encapsulating
        the undefined mid-level attribute ``"some_submodule"``. :func:`eval`
        then attempts to access the ``"SomeType[T]"`` attribute of the forward
        reference proxy for ``"some_submodule"``. Doing so implicitly invokes
        this dunder method, which then instantiates a final forward reference
        proxy encapsulating the undefined leaf-level attribute
        ``"SomeType[T]"``. Lastly, :func:`eval` then evaluates the expression
        ``"some_package.some_submodule.SomeType[T]"`` to the proxy for
        ``"SomeType[T]"``, which :func:`eval` then returns as its value. The
        intermediate proxies for both ``"some_package"`` and
        ``"some_submodule"`` are now irrelevant and thus garbage-collectable.

        This scheme is simple, effective, and (most importantly) efficient. But
        it's also prone to overly permissive proxying that intersects poorly
        with the second use case. Why? Because this scheme naively assumes that
        each invocation of this dunder method is **trustworthy**: that is, that
        each invocation of this dunder method is an attempt to access some valid
        module attribute that is known *a priori* to exist. This assumption
        holds for stringified absolute forward references used as type hints by
        the caller internally proxied by :mod:`beartype`. This assumption breaks
        when an invocation of this dunder method is **untrustworthy**: that is,
        when an invocation of this dunder method is merely an attempt to decide
        whether a module contains an attribute that is not known *a priori* to
        exist. In short, the second use case is the :func:`hasattr` builtin.

        The :func:`hasattr` builtin is actually implemented in terms of the
        :func:`getattr` builtin via the Easier to Ask for Permission than
        Forgiveness (EAFP) principle, implying that :func:`hasattr` internally
        invokes this dunder method. Although implemented in low-level C, a
        pure-Python implementation of :func:`hasattr` might vaguely resemble:

        .. code-block:: python

           def hasattr(obj: object, attr_name: str) -> bool:
               try:
                   getattr(obj, attr_name)
               except AttributeError:
                   return False
               return True

        The :func:`hasattr` builtin thus expects this dunder method to raise the
        :exc:`AttributeError` exception when the module proxied by this forward
        reference proxy fails to define an attribute with the passed name.
        However, this expectation conflicts with the overly permissive proxying
        performed by the scheme outlined above. In that first use case,
        :mod:`beartype` encapsulates an external module that is *not* safely
        importable with this forward reference proxy. Since :mod:`beartype` has
        *no* safe means of deciding whether that module actually defines an
        attribute with the passed name or not, :mod:`beartype` naively assumes
        that module to define that attribute. Under the scheme outlined above,
        this dunder method would *never* raise the :exc:`AttributeError` and the
        :func:`hasattr` attribute would *always* return :data:`True` when passed
        a forward reference proxy.

        Does this second use case arise in practice? In theory, it shouldn't.
        After all, forward reference proxies are mostly isolated to private
        subpackages in the :mod:`beartype` codebase... *mostly.* In practice,
        this second use case commonly arises. For efficiency, :mod:`beartype`
        replaces unusable stringified absolute forward references that are root
        type hints annotating the parameters and returns of
        :func:`beartype.beartype`-decorated callable with usable forward
        reference proxies. Popular third-party frameworks like ``pytest`` and
        Django then introspect those forward reference proxies during their
        non-trivial workloads. This introspection either directly calls the
        :func:`hasattr` builtin *or* replicates that builtin in pure-Python to
        detect whether those forward reference proxies define framework-specific
        dunder attributes of relevance to those frameworks.

        Parameters
        ----------
        cls : Type[BeartypeForwardRefABC]
            Forward reference subclass to concatenate this basename against.
        hint_name : str
            Unqualified basename to be concatenated against this forward
            reference subclass.

        Returns
        -------
        BeartypeForwardRef
            Fully-qualified forward reference subclass concatenated as above.
        '''

        #FIXME: Unit test up this edge case, please.
        # If this forward reference proxy has already been resolved to its
        # referent (e.g., by a prior isinstance() or issubclass() check),
        # forward this dunder method call directly to that referent.
        if _is_forwardref_resolved(cls):
            return getattr(cls.__type_beartype__, hint_name)
        # Else, this forward reference proxy has yet to be resolved.
        #
        # If a non-existent dunder attribute was requested, assume this
        # erroneous attempt to access a non-existent attribute of this forward
        # reference proxy to *ACTUALLY* be an Easier to Ask for Permission than
        # Forgiveness (EAFP)-driven to detect whether this forward scope defines
        # this attribute ala the hasattr() builtin. See also the "Design"
        # subsection of this dunder method's docstring for further commentary.
        elif is_dunder(hint_name):
            # Raise the standard "AttributeError" exception expected by EAFP.
            #
            # Note that we intentionally avoid suffixing the exception message
            # by a "." character here. Why? Because Python treats
            # "AttributeError" exceptions as special. Notably, Python appears to
            # actually:
            # 1. Parse apart the messages of these exceptions for the
            #    double-quoted attribute name embedded in these messages.
            # 2. Suffix these messages by a "." character followed by a sentence
            #    suggesting an existing attribute with a similar name to that of
            #    the attribute name previously parsed from these messages.
            #
            # For example, given an erroneous lookup of a non-existent dunder
            # attribute "__nomnom_beartype__", Python expands the exception
            # message raised below into:
            #     AttributeError: Forward reference proxy "MuhRef" dunder
            #     attribute "__nomnom_beartype__" not found. Did you mean:
            #     '__name_beartype__'?
            raise AttributeError(
                f'Forward reference proxy "{cls.__name__}" '
                f'dunder attribute "{hint_name}" not found'
            )
        # Else, the caller resides inside the "beartype" package and is
        # requesting a non-existent non-dunder attribute. In this case, safely
        # assume this request to comprise a higher-level attempt to resolve an
        # absolute stringified forward reference (e.g., the request for the
        # "some_submodule" attribute from the "some_package" forward reference
        # proxy given the initial absolute stringified forward reference
        # "some_package.some_submodule.SomeType").

        # Avoid circular import dependencies.
        from beartype._check.forward.reference.fwdrefmake import (
            make_forwardref_indexable_subtype)

        # Return a new fully-qualified forward reference subclass concatenated
        # as described above.
        return make_forwardref_indexable_subtype(
            cls.__scope_name_beartype__,  # type: ignore[arg-type]
            f'{cls.__name_beartype__}.{hint_name}',
        )


    def __instancecheck__(cls: BeartypeForwardRef, obj: object) -> bool:  # type: ignore[misc]
        '''
        :data:`True` only if the passed object is an instance of the external
        class referenced by the passed **forward reference subclass** (i.e.,
        :class:`.BeartypeForwardRefABC` subclass whose metaclass is this
        metaclass and whose :attr:`.BeartypeForwardRefABC.__name_beartype__`
        class variable is the fully-qualified name of that external class).

        Parameters
        ----------
        cls : Type[BeartypeForwardRefABC]
            Forward reference subclass to test this object against.
        obj : object
            Arbitrary object to be tested as an instance of the external class
            referenced by this forward reference subclass.

        Returns
        -------
        bool
            :data:`True` only if this object is an instance of the external
            class referenced by this forward reference subclass.
        '''

        # Return true only if this forward reference subclass insists that this
        # object satisfies the external class referenced by this subclass.
        return cls.__is_instance_beartype__(obj)


    def __subclasscheck__(cls: BeartypeForwardRef, obj: object) -> bool:  # type: ignore[misc]
        '''
        :data:`True` only if the passed object is a subclass of the external
        class referenced by the passed **forward reference subclass** (i.e.,
        :class:`.BeartypeForwardRefABC` subclass whose metaclass is this
        metaclass and whose :attr:`.BeartypeForwardRefABC.__name_beartype__`
        class variable is the fully-qualified name of that external class).

        Parameters
        ----------
        cls : Type[BeartypeForwardRefABC]
            Forward reference subclass to test this object against.
        obj : object
            Arbitrary object to be tested as a subclass of the external class
            referenced by this forward reference subclass.

        Returns
        -------
        bool
            :data:`True` only if this object is a subclass of the external class
            referenced by this forward reference subclass.
        '''

        # Return true only if this forward reference subclass insists that this
        # object is an instance of the external class referenced by this
        # subclass.
        return cls.__is_subclass_beartype__(obj)


    def __repr__(cls: BeartypeForwardRef) -> str:  # type: ignore[misc]
        '''
        Machine-readable string representing this forward reference subclass.
        '''

        # Machine-readable representation to be returned.
        #
        # Note that this representation is intentionally prefixed by the
        # @beartype-specific substring "<forwardref ", resembling the
        # representation of classes (e.g., "<class 'bool'>"). Why? Because
        # various other @beartype submodules ignore objects whose
        # representations are prefixed by the "<" character, which are usefully
        # treated as having a standard representation that is ignorable for most
        # intents and purposes. This includes:
        # * The die_if_hint_pep604_inconsistent() raiser.
        cls_repr = (
            f'<forwardref {cls.__name__}('
              f'__name_beartype__={repr(cls.__name_beartype__)}'
            f', __scope_name_beartype__={repr(cls.__scope_name_beartype__)}'
        )

        #FIXME: Unit test this edge case, please.
        # If this is a subscripted forward reference subclass, append additional
        # metadata representing this subscription.
        #
        # Ideally, we would test whether this is a subclass of the
        # "_BeartypeForwardRefIndexedABC" superclass as follows:
        #     if issubclass(cls, _BeartypeForwardRefIndexedABC):
        #
        # Sadly, doing so invokes the __subclasscheck__() dunder method defined
        # above, which invokes the
        # BeartypeForwardRefABC.__is_subclass_beartype__() method defined
        # above, which tests the type referred to by this subclass rather than
        # this subclass itself. In short, this is why you play with madness.
        try:
            cls_repr += (
                f', __args_beartype__={repr(cls.__args_beartype__)}'
                f', __kwargs_beartype__={repr(cls.__kwargs_beartype__)}'
            )
        # If doing so fails with the expected "AttributeError", then this is
        # *NOT* a subscripted forward reference subclass. Since this is
        # ignorable, silently ignore this common case. *sigh*
        except AttributeError:
            pass

        # Close this representation.
        cls_repr += ')>'

        # Return this representation.
        return cls_repr

    # ....................{ PROPERTIES                     }....................
    @property
    def __type_beartype__(cls: BeartypeForwardRef) -> type:  # type: ignore[misc]
        '''
        **Forward referent** (i.e., type hint referenced by this forward
        reference subclass, which is usually but *not* necessarily a class).

        This class property is manually memoized for efficiency. However, note
        this class property is *not* automatically memoized (e.g., by the
        ``property_cached`` decorator). Why? Because manual memoization enables
        other functionality in the beartype codebase to explicitly unmemoize all
        previously memoized forward referents across all forward reference
        proxies, effectively forcing all subsequent calls of this property
        across all forward reference proxies to reimport their forward referents.
        Why is that desirable? Because other functionality in the beartype
        codebase detects when the user has manually reloaded user-defined
        modules defining user-defined types annotating user-defined callables
        previously decorated by the :mod:`beartype.beartype` decorator. Since
        reloading those modules redefines those types, all previously cached
        types (including those memoized by this property) *must* then be assumed
        to be invalid and thus uncached. In short, manual memoization allows
        beartype to avoid desynchronization between memoized and actual types.

        This class property is officially in the public :mod:`beartype` API and
        guaranteed to be available across *all* current and future
        :mod:`beartype` releases.

        Caveats
        -------
        Downstream callers consuming callable type hints modified by a
        previously applied :mod:`beartype.beartype` decorator may occasionally
        encounter **forward reference proxies** (i.e., instances of this
        metaclass). Forward reference proxies are *not* intended to be usable as
        perfect substitutes for the underlying classes they proxy. Instead,
        downstream callers are recommended to manually resolve these proxies to
        the underlying classes they proxy by accessing this property. Consider
        this trivial one-liner that does so for a type hint ``type_hint``:

        .. code-block:: python

           # If this type hint is actually a @beartype-specific forward
           # reference proxy that only refers to the desired type hint,
           # dereference that proxy to obtain that type hint.
           type_hint = getattr(type_hint, '__type_beartype__', type_hint)

        Raises
        ------
        BeartypeCallHintForwardRefException
            If either:

            * This forward referent is unimportable.
            * This forward referent is importable but either:

              * Not a type.
              * A type that is this forward reference proxy, implying this proxy
                circularly proxies itself.
        '''

        # Forward referent referred to by this forward reference proxy if a prior
        # access of this property has already resolved this referent *OR* "None"
        # otherwise (i.e., if this is the first access of this property).
        referent = _forwardref_to_referent_get(cls)

        # If this forward referent has yet to be resolved, this is the first call
        # to this property. In this case...
        if referent is None:  # type: ignore[has-type]
            # print(f'Importing forward ref "{cls.__name_beartype__}" from module "{cls.__scope_name_beartype__}"...')

            # Exception subclass and prefix to be raised below.
            EXCEPTION_CLS = BeartypeCallHintForwardRefException
            EXCEPTION_PREFIX = 'Forward reference '

            # Forward referent dynamically imported from this module.
            referent = import_module_attr(
                attr_name=cls.__name_beartype__,
                module_name=cls.__scope_name_beartype__,
                exception_cls=EXCEPTION_CLS,
                exception_prefix=EXCEPTION_PREFIX,
            )

            # If this referent is this forward reference subclass, then this
            # subclass circularly proxies itself. Since allowing this edge case
            # would openly invite infinite recursion, we detect this edge case
            # and instead raise a human-readable exception.
            if referent is cls:
                raise BeartypeCallHintForwardRefException(
                    f'Forward reference proxy {repr(cls)} circularly '
                    f'(i.e., infinitely recursively) references itself.'
                )
            # Else, this referent is *NOT* this forward reference subclass.
            #
            # If this referent is a subscripted generic (e.g.,
            # "MuhGeneric[int]"), reduce this referent to the class subscripting
            # this generic (e.g., "int").
            elif is_hint_pep484585_generic(referent):
                referent = get_hint_pep484585_generic_type(
                    hint=referent,
                    exception_cls=EXCEPTION_CLS,
                    exception_prefix=EXCEPTION_PREFIX,
                )
            # Else, this referent is *NOT* a subscripted generic.

            # Cache this referent for subsequent lookup by this property
            # *BEFORE* validating this referent to be isinstanceable. If this
            # property is validated to *NOT* be isinstanceable, this referent
            # will be immediately uncached. Of course, this is insane. Ideally,
            # this referent would be cached only *AFTER* validating this
            # referent to be isinstanceable. Pragmatically, doing so invites
            # infinite recursion as follows (in order):
            # * This __type_beartype__() property getter calls...
            # * die_unless_object_isinstanceable(), which calls...
            # * "isinstance(None, cls)", which calls...
            # * BeartypeForwardRefMeta.__subclasscheck__(), which calls...
            # * "issubclass(obj, cls.__type_beartype__)", which calls...
            # * This __type_beartype__() property getter, which calls...
            # * die_unless_object_isinstanceable(). Repeat as needed for pain.
            #
            # Caching this referent first circumvents this recursion by ensuring
            # that all subsequent access of this property after the first access
            # of this property casually returns this referent rather than
            # repeatedly (thus uselessly) calling the
            # die_unless_object_isinstanceable() validator.
            _forwardref_to_referent[cls] = referent

            #FIXME: *SUPER-AWKWARD.* Slow, too. Ideally, we should instead:
            #* Define a new is_object_isinstanceable() tester. Note that this
            #  will be somewhat non-trivial (well -- tedious, mostly), which is
            #  why we haven't bothered yet. *sigh*
            #* Refactor the following "try: ... except Exception:" logic as follows:
            #     if not is_object_isinstanceable(referent):
            #         del _forwardref_to_referent[cls]
            #         die_unless_object_isinstanceable(
            #             obj=referent,
            #             exception_cls=BeartypeCallHintForwardRefException,
            #             exception_prefix='Forward reference ',
            #         )

            # Attempt to...
            try:
                # If this referent is *NOT* isinstanceable, raise an exception.
                die_unless_object_isinstanceable(
                    obj=referent,
                    exception_cls=EXCEPTION_CLS,
                    exception_prefix=EXCEPTION_PREFIX,

                    # If this referent is itself a forward reference proxy,
                    # raise an exception if that proxy *CANNOT* be resolved to
                    # the referent that proxy refers to. While an unlikely edge
                    # case, unlikely edge cases are like million-to-one chances
                    # in a Pratchett novel: you just know it's coming up.
                    is_forwardref_valid=False,
                )
                # Else, this referent is isinstanceable.
            # If doing so raised *ANY* exception whatsoever...
            except Exception:
                # Uncache this referent. See above.
                del _forwardref_to_referent[cls]

                # Re-raise this exception as is.
                raise
        # Else, this referent has already been resolved.
        #
        # In either case, this referent is now resolved.

        # Return this previously resolved referent.
        return referent  # type: ignore[return-value]

# ....................{ PRIVATE ~ globals                  }....................
_forwardref_to_referent: Dict[BeartypeForwardRef, type] = {}
'''
**Forward reference referent cache** (i.e., dictionary mapping from each forward
reference proxy to the arbitrary class referred to by that proxy).

This cache serves a dual purpose. Notably, this cache both enables:

* External callers to iterate over all previously instantiated forward reference
  proxies. This is particularly useful when responding to module reloading,
  which requires that *all* previously cached types be uncached.
* The
  :attr:`.BeartypeForwardRefMeta.__type_beartype__` property to internally
  memoize the arbitrary class referred to by this referent. Since the existing
  ``property_cached`` decorator could trivially do so as well, however, this is
  only a negligible side effect.
'''


_forwardref_to_referent_get = _forwardref_to_referent.get
'''
:meth:`dict.get` method bound to the :data:`._forwardref_to_referent` global
dictionary, globalized as a negligible microoptimization.
'''

# ....................{ PRIVATE ~ globals                  }....................
#FIXME: Unit test us up, please.
def _is_forwardref_resolved(hint: BeartypeForwardRef) -> bool:
    '''
    :data:`True` only if the passed **forward reference proxy** (i.e.,
    :class:`BeartypeForwardRef` object) is already been resolved to its
    **referent** (i.e., the object referred to by this reference).

    Parameters
    ----------
    hint : BeartypeForwardRef
        Forward reference proxy to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this proxy has been resolved to its referent.
    '''

    # Return true only if this proxy has been resolved to its referent.
    return hint in _forwardref_to_referent



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **TensorDict utilities** (i.e., low-level callables handling the
third-party :mod:`tensordict` package).

This private submodule is *not* intended for importation by downstream callers.
'''

#FIXME: Preserved for posterity. It's unclear whether @beartype currently
#requires (or desires) explicit TensorDict support. Ideally, it doesn't!

# # ....................{ IMPORTS                            }....................
# #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# # CAUTION: The top-level of this module should avoid importing from third-party
# # optional libraries, both because those libraries cannot be guaranteed to be
# # either installed or importable here *AND* because those imports are likely to
# # be computationally expensive, particularly for imports transitively importing
# # C extensions (e.g., anything from NumPy or SciPy).
# #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# # from dataclasses import is_dataclass
#
# # ....................{ TESTERS                            }....................
# #FIXME: Unit test us up, please.
# def is_tensorclass(obj: object) -> bool:
#     '''
#     :data:`True` only if the passed object is a **tensorclass** (i.e.,
#     :obj:`tensordict.tensorclass`-decorated type).
#
#     Tensorclasses currently violate :func:`beartype.beartype`-based
#     type-checking. Consequently, :mod:`beartype` *must* explicitly ignore and
#     thus detect tensorclasses. However, tensorclasses lack an identifiable
#     metaclass or superclass! Ergo, even detecting tensorclasses is non-trivial.
#     Although the :func:`tensordict.is_tensorclass` public function appears to
#     superficially serve a similar purpose as this homegrown tester, calling the
#     former would effectively require importing the extremely costly
#     :mod:`tensordict` package from a global scope in :mod:`beartype`! Ergo, we
#     intentionally define this competing alternative instead -- which does *not*
#     require importing that package and thus remains highly efficient.
#
#     Parameters
#     ----------
#     obj: object
#         Object to be inspected.
#
#     See Also
#     --------
#     https://github.com/beartype/beartype/issues/501
#         Upstream issue strongly inspiring this implementation.
#     '''
#
#     # Return true only if...
#     return (
#         #FIXME: Does this actually work? If so, this is preferable:
#         # is_dataclass(obj) and
#
#         # This object is a type *AND*...
#         isinstance(obj, type) and
#
#         # The type of this type is the "type" superclass (implying this type to
#         # have *NO* metaclass) *AND*...
#         obj.__class__ is type and
#
#         #FIXME: What does the tensordict.is_tensorclass() tester do? This?
#         #Something else? Whatever that tester does, we should do maybe too.
#         # This class defines the @tensorclass-specific private "_is_tensorclass"
#         # attribute. While fragile, the inadequate @tensorclass API leaves us
#         # little leg room. Fragility is the best we can currently do.
#         hasattr(obj, '_is_tensorclass')
#     )
